#!/usr/bin/env python3
"""
Test `corrct.projectors` package.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.spatial.transform import Rotation
import pytest
import skimage.data as skd
import skimage.transform as skt
from numpy.typing import NDArray
import corrct as cct


phantom_sl2d = skd.shepp_logan_phantom()
phantom_nuc3d, background_nuc3d, _ = cct.testing.create_phantom_nuclei3d(FoV_size=100)

ITERATIONS = 100

NUM_PHOTONS = 1e1
BACKGROUND_AVG = 2e0
ADD_POISSON = True


def _get_angles(angles_num: int = 41, angles_start: float = 0.0, angles_range: float = 180.0) -> NDArray:
    angles_deg = np.linspace(angles_start, angles_start + angles_range, angles_num, endpoint=True)
    return np.deg2rad(angles_deg)


def _get_shifts(angles_num: int, theo_rot_axis: float = 0.0) -> NDArray:
    sigma_error = 0.25
    linear_error = -0.05
    exponential_error = 7.5

    random_drifts = sigma_error * np.random.randn(angles_num)
    linear_drifts = linear_error * np.linspace(-(angles_num - 1) / 2, (angles_num - 1) / 2, angles_num)
    exponential_drifts = exponential_error * np.exp(-np.linspace(0, 5, angles_num))

    theo_shifts = random_drifts + linear_drifts + exponential_drifts + theo_rot_axis
    return np.around(theo_shifts, decimals=2)


def _generate_noise(data_theo: NDArray) -> NDArray:
    data_noise, data_theo, background = cct.testing.add_noise(
        data_theo, num_photons=NUM_PHOTONS, add_poisson=ADD_POISSON, background_avg=BACKGROUND_AVG
    )

    return data_noise - background


def test_api_det_shifts_vu():
    """Test the API of the detector VU shifts manipulation in `models`."""
    num_angles: int = 41
    angles_deg = np.linspace(0, 180, num_angles)
    shifts_u = np.random.randn(num_angles)
    shifts_v = np.random.randn(num_angles)

    shifts_vu = cct.models.combine_shifts_vu(shifts_v, shifts_u)
    assert np.all(shifts_vu[-1] == shifts_u), "Wrong composition of shifts from `cct.models.combine_shifts_vu`"
    assert np.all(shifts_vu[-2] == shifts_v), "Wrong composition of shifts from `cct.models.combine_shifts_vu`"

    prj_geom = cct.models.get_prj_geom_parallel(geom_type="2d")
    prj_geom.set_detector_shifts_vu(shifts_u)

    msg = f"Wrong dimensions of `det_pos_xyz` in ProjectionGeometry. Should be 2-dimensional, of shape: {[num_angles, 3]}"
    assert prj_geom.det_pos_xyz.ndim == 2, msg
    assert np.all(np.array(prj_geom.det_pos_xyz.shape) == [num_angles, 3]), msg
    assert np.all(prj_geom.det_pos_xyz[:, -3] == shifts_u), "The shifts along U are wrong"

    rot_prj_geom = prj_geom.rotate(np.deg2rad(10))
    assert rot_prj_geom.det_pos_xyz.ndim == 2, msg
    assert np.all(np.array(rot_prj_geom.det_pos_xyz.shape) == [num_angles, 3]), msg

    rot_prj_geom = prj_geom.rotate(np.deg2rad(angles_deg))
    assert rot_prj_geom.det_pos_xyz.ndim == 2, msg
    assert np.all(np.array(rot_prj_geom.det_pos_xyz.shape) == [num_angles, 3]), msg

    prj_geom = cct.models.get_prj_geom_parallel(geom_type="3d")
    prj_geom.set_detector_shifts_vu(shifts_vu)

    msg = f"Detector position in ProjectionGeometry should be 2-dimensional, of shape: {[num_angles, 3]}"
    assert prj_geom.det_pos_xyz.ndim == 2, msg
    assert np.all(np.array(prj_geom.det_pos_xyz.shape) == [num_angles, 3]), msg
    assert np.all(prj_geom.det_pos_xyz[:, -3] == shifts_u), "The shifts along U are wrong"
    assert np.all(prj_geom.det_pos_xyz[:, -1] == shifts_v), "The shifts along v are wrong"

    rot_prj_geom = prj_geom.rotate(np.deg2rad(10))
    assert rot_prj_geom.det_pos_xyz.ndim == 2, msg
    assert np.all(np.array(rot_prj_geom.det_pos_xyz.shape) == [num_angles, 3]), msg

    rot_prj_geom = prj_geom.rotate(np.deg2rad(angles_deg))
    assert rot_prj_geom.det_pos_xyz.ndim == 2, msg
    assert np.all(np.array(rot_prj_geom.det_pos_xyz.shape) == [num_angles, 3]), msg

    prj_geom = cct.models.get_prj_geom_parallel(geom_type="3d")
    prj_geom.set_detector_shifts_vu(shifts_vu)

    prj_geom.set_detector_tilt(np.pi / 2)

    assert np.all(np.isclose(prj_geom.det_pos_xyz[:, -1], -shifts_u)), "The tilted shifts along U are wrong"
    assert np.all(np.isclose(prj_geom.det_pos_xyz[:, -3], shifts_v)), "The tilted shifts along v are wrong"

    prj_geom.set_detector_tilt(np.pi / 2)

    assert np.all(np.isclose(prj_geom.det_pos_xyz[:, -3], -shifts_u)), "The tilted shifts along U are wrong"
    assert np.all(np.isclose(prj_geom.det_pos_xyz[:, -1], -shifts_v)), "The tilted shifts along v are wrong"


def test_set_detector_tilt():
    """
    Test ProjectionGeometry.set_detector_tilt method with various scenarios of angles, axes and source tilting.

    Checks if detector u and v vectors are correctly tilted after calling the function. Also checks if the source is
    correctly tilted when specified.
    """
    # Setup: Create a default geometry
    prj_geom_orig = cct.models.get_prj_geom_parallel(geom_type="3d")

    prj_geom_test = prj_geom_orig.copy()
    # Test 1: Single tilt along the default axis (0, 1, 0)
    angles_t_rad = np.deg2rad(30)
    prj_geom_test.set_detector_tilt(angles_t_rad)

    rotation = Rotation.from_rotvec(angles_t_rad * np.array([0, 1, 0]))
    expected_u = rotation.apply(prj_geom_orig.det_u_xyz)
    assert np.allclose(prj_geom_test.det_u_xyz, expected_u), "Failed: Single tilt along default axis"

    prj_geom_test = prj_geom_orig.copy()
    # Test 2: Single tilt along a custom axis
    angles_t_rad = np.deg2rad(-45)
    tilt_axis = (1, 0, 0)
    prj_geom_test.set_detector_tilt(angles_t_rad, tilt_axis=tilt_axis)

    rotation = Rotation.from_rotvec(angles_t_rad * np.array(tilt_axis))
    expected_v = rotation.apply(prj_geom_orig.det_v_xyz)
    assert np.allclose(prj_geom_test.det_v_xyz, expected_v), "Failed: Single tilt along custom axis"

    prj_geom_test = prj_geom_orig.copy()
    # Test 3: Multiple tilts along different axes
    angles_t_rad = np.deg2rad([45, -60])
    tilt_axis = [(1, 0, 0), (0, 1, 0)]
    prj_geom_test.set_detector_tilt(angles_t_rad, tilt_axis=tilt_axis)

    expected_u = prj_geom_orig.det_u_xyz
    for angle, axis in zip(angles_t_rad, tilt_axis):
        rotations = Rotation.from_rotvec(angle * np.array(axis))  # type: ignore
        expected_u = rotations.apply(expected_u)

    assert np.allclose(prj_geom_test.det_u_xyz, expected_u), "Failed: Multiple tilts along different axes"

    prj_geom_test = prj_geom_orig.copy()
    # Test 4: Tilt the source as well
    angles_t_rad = np.deg2rad(-90)
    tilt_source = True
    prj_geom_test.set_detector_tilt(angles_t_rad, tilt_source=tilt_source)

    rotation = Rotation.from_rotvec(angles_t_rad * np.array([0, 1, 0]))
    expected_src = rotation.apply(prj_geom_orig.src_pos_xyz)
    assert np.allclose(prj_geom_test.src_pos_xyz, expected_src), "Failed: Tilting the source"


def test_fit_image_rotation_and_scale():
    """
    Test the fit_image_rotation_and_scale function from the cct.alignment.fitting module.

    This function should return a tuple (angle, scale) that represents the rotation angle
    and scaling factor required to align an input image with a reference image. The
    function is tested using various combinations of identical images, simple rotations,
    scaling, and both rotation and scaling. It is also tested with images of different
    shapes, in which case it should raise a ValueError.
    """
    # Load the camera image from scikit-image's data package
    img: NDArray = skd.camera() / 255

    # Test with two identical images should return 0 degrees and 1 scale
    result = cct.alignment.fitting.fit_image_rotation_and_scale(img, img)
    assert np.allclose(result, (0.0, 1.0))

    # Test with a simple rotation of an image by 10 degrees
    img_10 = skt.rotate(img, angle=10.0)
    result = cct.alignment.fitting.fit_image_rotation_and_scale(img, img_10)
    assert np.allclose(result, (10.0, 1.0), rtol=0.01)

    # Test with a simple rotation of an image by 45 degrees
    img_45 = skt.rotate(img, angle=45.0)
    result = cct.alignment.fitting.fit_image_rotation_and_scale(img, img_45)
    assert np.allclose(result, (45.0, 1.0), rtol=0.01)

    # Test with a rotation of an image by 360 degrees
    img_360 = np.copy(img)
    result = cct.alignment.fitting.fit_image_rotation_and_scale(img, img_360)
    assert np.allclose(result, (0.0, 1.0), rtol=0.01)

    # Test with a scaling of an image by 10% using scikit-image's rescale function
    img_scaled: NDArray = skt.rescale(img, scale=1.1)
    center_y, center_x = img.shape[0] // 2, img.shape[1] // 2
    img_scaled_cropped = img_scaled[
        center_y - img.shape[0] // 2 : center_y + img.shape[0] // 2,
        center_x - img.shape[1] // 2 : center_x + img.shape[1] // 2,
    ]
    result = cct.alignment.fitting.fit_image_rotation_and_scale(img, img_scaled_cropped)
    assert np.allclose(result, (0.0, 1.1), rtol=0.01, atol=0.05)

    # Test with a rotation and scaling of an image
    img_scaled = skt.rescale(img, scale=1.1)
    img_scaled_cropped = img_scaled[
        center_y - img.shape[0] // 2 : center_y + img.shape[0] // 2,
        center_x - img.shape[1] // 2 : center_x + img.shape[1] // 2,
    ]
    img_10_scaled = skt.rotate(img_scaled_cropped, angle=10.0)
    result = cct.alignment.fitting.fit_image_rotation_and_scale(img, img_10_scaled)
    assert np.allclose(result, (10.0, 1.1), rtol=0.01)

    # Test with images of different shapes should raise a ValueError
    try:
        cct.alignment.fitting.fit_image_rotation_and_scale(img, np.random.rand(10, 20))
        assert False, "Expected a ValueError"
    except ValueError:
        pass


@pytest.mark.skipif(not cct.projectors.astra_available, reason="astra-toolbox not available")
@pytest.mark.parametrize("add_noise", [(False,), (True,)])
def test_pre_alignment(add_noise: bool, theo_rot_axis: float = -1.25):
    """Test pre-alignment routines.

    Parameters
    ----------
    add_noise : bool
        Whether to add noise
    theo_rot_axis : float, optional
        The theoretical rotation axis position, by default -1.25
    """
    debug = False

    vol_geom = cct.models.get_vol_geom_from_volume(phantom_sl2d)
    angles_rad = _get_angles()
    theo_shifts = _get_shifts(len(angles_rad), theo_rot_axis=theo_rot_axis)

    with cct.projectors.ProjectorUncorrected(vol_geom, angles_rad, theo_shifts) as prj:
        prj_data = prj(phantom_sl2d)

    if add_noise:
        prj_data = _generate_noise(prj_data)

    # Setting up the pre-alignment routine
    align_pre = cct.alignment.DetectorShiftsPRE(prj_data, angles_rad, verbose=debug)

    # Running pre-alignment
    diffs_u_pre, cor = align_pre.fit_u()
    shifts_u_pre = cor + diffs_u_pre

    solver_opts = dict(lower_limit=0.0)

    solver = cct.solvers.SIRT()
    if debug:
        with cct.projectors.ProjectorUncorrected(vol_geom, angles_rad, theo_rot_axis) as prj:
            rec_noise_theocor, _ = solver(prj, prj_data, iterations=ITERATIONS, **solver_opts)
        with cct.projectors.ProjectorUncorrected(vol_geom, angles_rad, cor) as prj:
            rec_noise_precor, _ = solver(prj, prj_data, iterations=ITERATIONS, **solver_opts)
    with cct.projectors.ProjectorUncorrected(vol_geom, angles_rad, shifts_u_pre) as prj:
        rec_noise_pre, _ = solver(prj, prj_data, iterations=ITERATIONS, **solver_opts)

    com_ph_yx = cct.processing.post.com(phantom_sl2d)
    prj_geom = cct.models.ProjectionGeometry.get_default_parallel(geom_type="2d")
    recenter = cct.alignment.RecenterVolume(prj_geom, angles_rad)

    # Re-centering the reconstructions on the phantom's center-of-mass -> moving the shifts accordingly
    shifts_u_pre = recenter.to_com(shifts_u_pre, rec_noise_pre, com_ph_yx)

    if debug:
        theo_rot_axis_per_angle = np.ones_like(angles_rad) * theo_rot_axis
        theo_rot_axis_per_angle = recenter.to_com(theo_rot_axis_per_angle, rec_noise_theocor, com_ph_yx)
        cor = recenter.to_com(cor, rec_noise_precor, com_ph_yx)

        print(f"{theo_shifts = }")
        print(f"{shifts_u_pre = }")
        print(f"{theo_shifts - shifts_u_pre = }")
        print(f"{np.max(np.abs(theo_shifts - shifts_u_pre)) = }")

        fig, axs = plt.subplots(1, 1, sharex=True, sharey=True, figsize=[5, 2.5])
        axs.plot(theo_shifts, label="Ground truth")
        axs.plot(theo_rot_axis_per_angle, label="Center-of-rotation (theoretical)")
        axs.plot(cor, label="Center-of-rotation (computed)")
        axs.plot(shifts_u_pre, label="Pre-alignment shifts")
        axs.grid()
        axs.legend()
        fig.tight_layout()
        plt.show()

    tolerance = 0.8 if add_noise else 0.3
    assert np.allclose(shifts_u_pre, theo_shifts, atol=tolerance), "Theoretical and computed shifts do not match"


@pytest.mark.skipif(not cct.projectors.astra_available, reason="astra-toolbox not available")
@pytest.mark.parametrize("add_noise", [(False,), (True,)])
def test_pre_alignment_3d(add_noise: bool, theo_rot_axis: float = -1.25):
    """Test pre-alignment routines.

    Parameters
    ----------
    add_noise : bool
        Whether to add noise
    theo_rot_axis : float, optional
        The theoretical rotation axis position, by default -1.25
    """
    debug = False

    phantom = np.fmax(phantom_nuc3d - background_nuc3d, 0.0)
    phantom /= phantom.max()

    vol_geom = cct.models.get_vol_geom_from_volume(phantom)
    angles_rad = _get_angles(angles_range=360)
    theo_shifts_u = _get_shifts(len(angles_rad), theo_rot_axis=theo_rot_axis) / 5.0
    theo_rot_axis = theo_rot_axis / 5.0
    theo_shifts_v = _get_shifts(len(angles_rad)) / 20.0
    theo_shifts_vu = cct.models.combine_shifts_vu(theo_shifts_v, theo_shifts_u)
    prj_geom = cct.models.get_prj_geom_parallel()
    prj_geom.set_detector_shifts_vu(theo_shifts_vu)

    with cct.projectors.ProjectorUncorrected(vol_geom, angles_rad, prj_geom=prj_geom) as prj:
        prj_data = prj(phantom)

    if debug:
        fig, axs = plt.subplots(1, 2, figsize=[8, 2.5])
        axs[0].imshow(phantom_nuc3d[phantom.shape[0] // 2])
        axs[0].set_title("Phantom + background")
        axs[1].imshow(background_nuc3d[phantom.shape[0] // 2])
        axs[1].set_title("Background")
        fig.tight_layout()

        fig, axs = plt.subplots(1, 2, figsize=[8, 2.5])
        axs[0].imshow(phantom[phantom.shape[0] // 2])
        axs[0].set_title("Phantom")
        axs[1].imshow(prj_data[phantom.shape[0] // 2])
        axs[1].set_title("Sinogram")
        fig.tight_layout()

    if add_noise:
        prj_data = _generate_noise(prj_data)

    # Setting up the pre-alignment routine
    align_pre = cct.alignment.DetectorShiftsPRE(prj_data, angles_rad, verbose=debug)

    # Running pre-alignment
    shifts_v_pre = align_pre.fit_v(use_derivative=False)
    shifts_u_pre, cor_pre = align_pre.fit_u(background=5.0)
    shifts_vu_pre = cct.models.combine_shifts_vu(shifts_v_pre, shifts_u_pre)
    prj_geom = cct.models.get_prj_geom_parallel()
    prj_geom.set_detector_shifts_vu(shifts_vu_pre, cor_pos_u=cor_pre)

    solver_opts = dict(lower_limit=0.0)

    solver = cct.solvers.SIRT()
    with cct.projectors.ProjectorUncorrected(vol_geom, angles_rad, theo_shifts_vu) as prj:
        rec_noise_theo, _ = solver(prj, prj_data, iterations=ITERATIONS, **solver_opts)
    with cct.projectors.ProjectorUncorrected(vol_geom, angles_rad, prj_geom=prj_geom) as prj:
        rec_noise_pre, _ = solver(prj, prj_data, iterations=ITERATIONS, **solver_opts)
    if debug:
        with cct.projectors.ProjectorUncorrected(vol_geom, angles_rad, theo_rot_axis) as prj:
            rec_noise_theocor, _ = solver(prj, prj_data, iterations=ITERATIONS, **solver_opts)
        with cct.projectors.ProjectorUncorrected(vol_geom, angles_rad, cor_pre) as prj:
            rec_noise_precor, _ = solver(prj, prj_data, iterations=ITERATIONS, **solver_opts)

    com_ph_yx = cct.processing.post.com(phantom)
    prj_geom = cct.models.get_prj_geom_parallel()
    recenter = cct.alignment.RecenterVolume(prj_geom, angles_rad)

    theo_shifts_vu = recenter.to_com(theo_shifts_vu, rec_noise_theo, com_ph_yx)
    shifts_vu_pre = recenter.to_com(shifts_vu_pre, rec_noise_pre, com_ph_yx)

    # Re-centering the reconstructions on the phantom's center-of-mass -> moving the shifts accordingly
    # shifts_vu_pre = recenter.to_com(shifts_vu_pre, rec_noise_xc, com_ph_yx)
    if debug:
        theo_rot_axis_per_angle = cct.models.combine_shifts_vu(
            np.zeros_like(angles_rad), np.ones_like(angles_rad) * theo_rot_axis
        )
        theo_rot_axis_per_angle = recenter.to_com(theo_rot_axis_per_angle, rec_noise_theocor, com_ph_yx)
        cor_pre = cct.models.combine_shifts_vu(np.zeros_like(angles_rad), np.ones_like(angles_rad) * cor_pre)
        cor_pre = recenter.to_com(cor_pre, rec_noise_precor, com_ph_yx)

        print(f"{theo_shifts_vu = }")
        print(f"{shifts_vu_pre = }")
        print(f"{theo_shifts_vu - shifts_vu_pre = }")
        print(f"{np.max(np.abs(theo_shifts_vu[0] - shifts_vu_pre[0])) = }")
        print(f"{np.max(np.abs(theo_shifts_vu[1] - shifts_vu_pre[1])) = }")

        fig, axs = plt.subplots(1, 2, sharex=True, sharey=True, figsize=[8, 2.5])
        axs[0].plot(theo_shifts_vu[0], label="Ground truth")
        axs[0].plot(shifts_vu_pre[0], label="Pre-alignment shifts")
        axs[0].grid()
        axs[0].legend()
        axs[0].set_title("Vertical shifts")
        axs[1].plot(theo_shifts_vu[1], label="Ground truth")
        axs[1].plot(theo_rot_axis_per_angle[1], label="Center-of-rotation (theoretical)")
        axs[1].plot(cor_pre[1], label="Center-of-rotation (computed)")
        axs[1].plot(shifts_vu_pre[1], label="Pre-alignment shifts")
        axs[1].plot(shifts_vu_pre[1] - theo_shifts_vu[1], label="Diff")
        axs[1].grid()
        axs[1].legend()
        axs[1].set_title("Horizontal shifts")
        fig.tight_layout()
        plt.show()

    tolerance = 0.8 if add_noise else 0.25
    assert np.allclose(
        shifts_vu_pre[0], theo_shifts_vu[0], atol=tolerance
    ), "Theoretical and computed vertical shifts do not match"
    tolerance = 0.8 if add_noise else 0.5
    assert np.allclose(
        shifts_vu_pre[1], theo_shifts_vu[1], atol=tolerance
    ), "Theoretical and computed horizontal shifts do not match"



#!/usr/bin/env python

"""Tests for `corrct.operators` package."""

import numpy as np
import pytest
import scipy.signal as spsig

from corrct import operators, testing

eps = float(np.finfo(np.float32).eps)


@pytest.fixture(scope="class")
def bootstrap_base(request):
    """Set up the testing classes."""
    cls = request.cls
    test_vols_shape = (29, 29)
    cls.vol_ones_2d = np.ones(test_vols_shape)


@pytest.mark.usefixtures("bootstrap_base")
class TestTransformGradient:
    """Test for TransformGradient class in `corrct.operators` package."""

    def test_000_gradient(self):
        """Test gradient operator 2D."""
        D = operators.TransformGradient(self.vol_ones_2d.shape, pad_mode="constant")

        g = D(self.vol_ones_2d)
        assert np.allclose(g.shape[1:], self.vol_ones_2d.shape)
        assert g.shape[0] == 2

        assert np.allclose(g[0, -1, :], -1, atol=eps)
        assert np.allclose(g[1, :, -1], -1, atol=eps)

    def test_001_minus_divergence(self):
        """Test divergence operator 2D."""
        D = operators.TransformGradient(self.vol_ones_2d.shape, pad_mode="constant")

        g = np.ones(D.adj_shape)
        d = D.T(g)
        assert np.allclose(d.shape, self.vol_ones_2d.shape)

        assert np.allclose(g[0, 0, :], 1, atol=eps)
        assert np.allclose(g[1, :, 0], 1, atol=eps)

    def test_002_explicit_gradient(self):
        """Test explicit gradient operator 2D."""
        D = operators.TransformGradient(self.vol_ones_2d.shape, pad_mode="constant")
        g0 = D(self.vol_ones_2d)

        De = D.explicit()
        ge = De.dot(self.vol_ones_2d.flatten())
        assert g0.size == ge.size

        ge = np.reshape(ge, D.adj_shape)
        assert np.allclose(g0, ge, atol=eps)

    def test_003_explicit_minus_divergence(self):
        """Test explicit transposed gradient operator 2D."""
        D = operators.TransformGradient(self.vol_ones_2d.shape, pad_mode="constant")
        g = np.ones(D.adj_shape)
        d0 = D.T(g)

        Dte = D.T.explicit()
        de = Dte.dot(np.ones(D.adj_shape).flatten())
        assert d0.size == de.size

        de = np.reshape(de, D.dir_shape)
        assert np.allclose(d0, de, atol=eps)


@pytest.mark.usefixtures("bootstrap_base")
class TestTransformLaplacian:
    """Test for TransformLaplacian class in `corrct.operators` package."""

    def test_000_laplacian(self):
        """Test laplacian operator 2D."""
        L = operators.TransformLaplacian(self.vol_ones_2d.shape)

        g = L(self.vol_ones_2d)
        assert np.allclose(g.shape, self.vol_ones_2d.shape)
        assert np.allclose(g, 0, atol=eps)

        test_line = np.ones((3,))
        test_line[1] = 0
        L = operators.TransformLaplacian(test_line.shape)
        g = L(test_line)
        assert np.allclose(g, [-1, 2, -1], atol=eps)

    def test_001_explicit_gradient(self):
        """Test explicit laplacian operator 2D."""
        L = operators.TransformLaplacian(self.vol_ones_2d.shape)
        g0 = L(self.vol_ones_2d)

        Le = L.explicit()
        ge = Le.dot(self.vol_ones_2d.flatten())
        assert g0.size == ge.size

        ge = np.reshape(ge, L.adj_shape)
        assert np.allclose(g0, ge, atol=eps)


@pytest.mark.usefixtures("bootstrap_base")
class TestTransformStationaryWavelet:
    """Tests for the TransformStationaryWavelet class in `corrct.operators` package."""

    def test_000_transform(self):
        """Test Haar wavelet transform in 2D."""
        wl_dec_level = 3
        H = operators.TransformStationaryWavelet(self.vol_ones_2d.shape, "db1", wl_dec_level)

        w = H(self.vol_ones_2d)
        assert np.allclose(w.shape[1:], testing.roundup_to_pow2(self.vol_ones_2d.shape, wl_dec_level))
        assert w.shape[0] == 10

        wtw = H.T(w)
        assert np.allclose(wtw.shape, self.vol_ones_2d.shape)

        print(f"Max absolute deviation is: {np.max(np.abs(self.vol_ones_2d - wtw))}. ", end="", flush=True)
        assert np.allclose(wtw, self.vol_ones_2d, atol=eps * 1e2)

    def test_001_explicit_transform(self):
        """Test explicit Haar transform in 2D."""
        H = operators.TransformStationaryWavelet(self.vol_ones_2d.shape, "db1", 3)
        w0 = H(self.vol_ones_2d)

        He = H.explicit()
        we = He.dot(self.vol_ones_2d.flatten())
        assert w0.size == we.size

        we = np.reshape(we, H.adj_shape)
        assert np.allclose(w0, we, atol=eps)

    def test_002_explicit_inverse_transform(self):
        """Test explicit inverse Haar transform in 2D."""
        H = operators.TransformStationaryWavelet(self.vol_ones_2d.shape, "db1", 2)
        wt0 = H.T(np.ones(H.adj_shape))

        Hte = H.T.explicit()
        wte = Hte.dot(np.ones(H.adj_shape).flatten())
        assert wt0.size == wte.size

        wte = np.reshape(wte, H.dir_shape)
        assert np.allclose(wt0, wte, atol=eps)


@pytest.fixture(scope="class")
def bootstrap_convolution(request):
    """Set up the convolution test class."""
    cls = request.cls

    cls.prj_shape_v1u = (32, 1, 32)
    cls.prj_shape_vu = (32, 32)
    cls.prj_shape_vwu = (32, 8, 32)

    cls.prj_shape_1u = cls.prj_shape_vwu[-2::]
    cls.prj_shape_u = cls.prj_shape_vwu[-1::]
    cls.prj_shape_wu = cls.prj_shape_vwu[-2::]

    cls.prj_vu = np.random.rand(*cls.prj_shape_vu)
    cls.prj_vwu = np.random.rand(*cls.prj_shape_vwu)

    cls.kernel_1u = np.ones((1, 3))
    cls.kernel_u = np.ones((3,))
    cls.kernel_v1u = np.ones((3, 1, 3))
    cls.kernel_vu = np.ones((3, 3))


@pytest.mark.usefixtures("bootstrap_convolution")
class TestTransformConvolution:
    """Tests for the TransformConvolution class in `corrct.operators` package."""

    def test_000_initialize(self):
        """Test the initialization of the convolution operator."""
        C = operators.TransformConvolution(self.prj_shape_vu, kernel=self.kernel_u)
        assert len(C.kernel.shape) == len(C.dir_shape), "Incorrect kernel dimensions initialization (img vu, ker u)"

        C = operators.TransformConvolution(self.prj_shape_vu, kernel=self.kernel_vu)
        assert len(C.kernel.shape) == len(C.dir_shape), "Incorrect kernel dimensions initialization (img vu, ker vu)"

        C = operators.TransformConvolution(self.prj_shape_vwu, kernel=self.kernel_u)
        assert len(C.kernel.shape) == len(C.dir_shape), "Incorrect kernel dimensions initialization (img vwu, ker u)"

        C = operators.TransformConvolution(self.prj_shape_vwu, kernel=self.kernel_vu)
        assert len(C.kernel.shape) == len(C.dir_shape), "Incorrect kernel dimensions initialization (img vwu, ker u)"

    def test_001_direct_shapes(self):
        """Test the output dimensions of the convolution."""
        C = operators.TransformConvolution(self.prj_shape_vu, kernel=self.kernel_u)
        conv_prj = C(self.prj_vu)
        assert len(conv_prj.shape) == len(self.prj_vu.shape), "Non-matching output dimensions (img vu, ker u)"

        C = operators.TransformConvolution(self.prj_shape_vu, kernel=self.kernel_vu)
        conv_prj = C(self.prj_vu)
        assert len(conv_prj.shape) == len(self.prj_vu.shape), "Non-matching output dimensions (img vu, ker vu)"

        C = operators.TransformConvolution(self.prj_shape_vwu, kernel=self.kernel_u)
        conv_prj = C(self.prj_vwu)
        assert len(conv_prj.shape) == len(self.prj_vwu.shape), "Non-matching output dimensions (img vwu, ker u)"

        C = operators.TransformConvolution(self.prj_shape_vwu, kernel=self.kernel_vu)
        conv_prj = C(self.prj_vwu)
        assert len(conv_prj.shape) == len(self.prj_vwu.shape), "Non-matching output dimensions (img vwu, ker vu)"

    def test_002_direct_results_vu_u(self):
        """Test the output correctness of the convolution for the case  (img vu, ker u)."""
        C = operators.TransformConvolution(self.prj_shape_vu, kernel=self.kernel_u, pad_mode="constant")
        conv_prj = C(self.prj_vu)
        conv_prj_ref = spsig.convolve(self.prj_vu, self.kernel_u[None, :], mode="same")
        assert np.allclose(conv_prj, conv_prj_ref, rtol=1e-7), "Non-matching output (img vu, ker u)"

    def test_002_direct_results_vu_vu(self):
        """Test the output correctness of the convolution for the case  (img vu, ker vu)."""
        C = operators.TransformConvolution(self.prj_shape_vu, kernel=self.kernel_vu, pad_mode="constant")
        conv_prj = C(self.prj_vu)
        conv_prj_ref = spsig.convolve(self.prj_vu, self.kernel_vu, mode="same")
        assert np.allclose(conv_prj, conv_prj_ref, rtol=1e-7), "Non-matching output (img vu, ker vu)"

    def test_002_direct_results_vwu_u(self):
        """Test the output correctness of the convolution for the case  (img vwu, ker u)."""
        C = operators.TransformConvolution(self.prj_shape_vwu, kernel=self.kernel_u, pad_mode="constant")
        conv_prj = C(self.prj_vwu)
        conv_prj_ref = spsig.convolve(self.prj_vwu, self.kernel_u[None, None, :], mode="same")
        assert np.allclose(conv_prj, conv_prj_ref, rtol=1e-7), "Non-matching output (img vwu, ker u)"

    def test_002_direct_results_vwu_vu(self):
        """Test the output correctness of the convolution for the case  (img vwu, ker vu)."""
        C = operators.TransformConvolution(self.prj_shape_vwu, kernel=self.kernel_vu[..., None, :], pad_mode="constant")
        conv_prj = C(self.prj_vwu)
        conv_prj_ref = spsig.convolve(self.prj_vwu, self.kernel_vu[..., None, :], mode="same")
        assert np.allclose(conv_prj, conv_prj_ref, rtol=1e-7), "Non-matching output (img vwu, ker vu)"



#!/usr/bin/env python3
"""
Operators module.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import copy as cp
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Callable, Optional, Union
import numpy as np
import scipy.signal as spsig
from numpy.typing import ArrayLike, NDArray
from scipy.sparse.linalg import LinearOperator


try:
    import pywt

    has_pywt = True
    use_swtn = pywt.version.version >= "1.0.2"  # type: ignore
    if not use_swtn:
        print("WARNING - pywavelets version is too old (<1.0.2)")
except ImportError:
    has_pywt = False
    use_swtn = False
    print("WARNING - pywt was not found")


NDArrayInt = NDArray[np.integer]


class BaseTransform(LinearOperator, ABC):
    """Base operator class.

    It implements the linear operator behavior that can be used with the solvers in the `.solvers` module,
    and by the solvers in `scipy.sparse.linalg`.

    Parameters
    ----------
    dir_shape : NDArrayInt
        Shape of the direct space.
    adj_shape : NDArrayInt
        Shape of the adjoint space

    Attributes
    ----------
    is_dir_operator : bool
        Flag indicating if the operator is a direct operator.

    Notes
    -----
        It assumes that the fields `dir_shape` and `adj_shape` have been set during the initialization of the derived classes.
    """

    dir_shape: NDArrayInt
    adj_shape: NDArrayInt

    def __init__(self) -> None:
        """Initialize the base operator class."""
        num_cols = np.prod(self.dir_shape)
        num_rows = np.prod(self.adj_shape)
        super().__init__(np.float32, [num_rows, num_cols])
        self.is_dir_operator = True

    def _matvec(self, x: NDArray) -> NDArray:
        """Implement the direct operator for column vectors from the right.

        Parameters
        ----------
        x : NDArray
            Either row from the left or column from the right.

        Returns
        -------
        NDArray
            Result of applying the direct operator.
        """
        if self.is_dir_operator:
            x = x.reshape(self.dir_shape)
            return self._op_direct(x).flatten()
        else:
            x = x.reshape(self.adj_shape)
            return self._op_adjoint(x).flatten()

    def rmatvec(self, x: NDArray) -> NDArray:
        """Implement the direct operator for row vectors from the left.

        Parameters
        ----------
        x : NDArray
            Either row from the left or column from the right on transpose.

        Returns
        -------
        NDArray
            Result of applying the direct operator for row vectors.
        """
        if self.is_dir_operator:
            x = x.reshape(self.adj_shape)
            return self._op_adjoint(x).flatten()
        else:
            x = x.reshape(self.dir_shape)
            return self._op_direct(x).flatten()

    def _transpose(self) -> "BaseTransform":
        """Create the transpose operator.

        Returns
        -------
        BaseTransform
            The transpose operator.
        """
        Op_t = cp.copy(self)
        Op_t.shape = [Op_t.shape[1], Op_t.shape[0]]
        Op_t.is_dir_operator = False
        return Op_t

    def _adjoint(self) -> "BaseTransform":
        """Create the adjoint operator.

        Returns
        -------
        BaseTransform
            The adjoint operator.
        """
        return self._transpose()

    def absolute(self) -> "BaseTransform":
        """Return the absolute value of the operator.

        Returns
        -------
        BaseTransform
            The absolute value operator.
        """
        return self

    def explicit(self) -> NDArray:
        """Return the explicit transformation matrix associated with the operator.

        Returns
        -------
        NDArray
            The explicit transformation matrix, as a NumPy array.
        """
        He = np.empty(self.shape, dtype=self.dtype)
        if self.is_dir_operator:
            dim_size = np.prod(self.dir_shape)
        else:
            dim_size = np.prod(self.adj_shape)
        for ii in range(dim_size):
            xii = np.zeros((dim_size,))
            xii[ii] = 1
            He[:, ii] = self * xii
        return He

    def __call__(self, x: NDArray) -> NDArray:
        """Apply the operator to the input vector `x`.

        Parameters
        ----------
        x : NDArray
            Input vector.

        Returns
        -------
        result : NDArray
            The result of applying the operator to `x`.
        """
        if self.is_dir_operator:
            return self._op_direct(x)
        else:
            return self._op_adjoint(x)

    @abstractmethod
    def _op_direct(self, x: NDArray) -> NDArray:
        """Apply the operator to the data.

        Parameters
        ----------
        x : NDArray
            Data to process.

        Returns
        -------
        NDArray
            The processed data.
        """

    @abstractmethod
    def _op_adjoint(self, x: NDArray) -> NDArray:
        """Apply the adjoint operator to the data.

        Parameters
        ----------
        x : NDArray
            Data to process.

        Returns
        -------
        NDArray
            The processed data.
        """


class TransformFunctions(BaseTransform):
    """Transform class that uses callables."""

    def __init__(
        self,
        dir_shape: Union[ArrayLike, NDArray],
        adj_shape: Union[ArrayLike, NDArray],
        A: Callable[[NDArray], NDArray],
        At: Optional[Callable[[NDArray], NDArray]] = None,
    ) -> None:
        """Initialize the callable transform.

        If the adjoint of the function is not given, the function is considered symmetric.

        Parameters
        ----------
        dir_shape : ArrayLike
            Shape of the direct space.
        adj_shape : ArrayLike
            Shape of the adjoint space.
        A : Callable[[NDArray], NDArray]
            The transform function.
        At : Optional[Callable[[NDArray], NDArray]], optional
            The adjoint transform function, by default None
        """
        self.dir_shape = np.array(dir_shape, ndmin=1, dtype=int)
        self.adj_shape = np.array(adj_shape, ndmin=1, dtype=int)
        self.A = A
        self.At = At
        super().__init__()

    def _op_direct(self, x: NDArray) -> NDArray:
        """Apply the operator to the data.

        Parameters
        ----------
        x : NDArray
            Data to process.

        Returns
        -------
        NDArray
            The processed data.
        """
        return self.A(x)

    def _op_adjoint(self, x: NDArray) -> NDArray:
        """Apply the adjoint operator to the data.

        Parameters
        ----------
        x : NDArray
            Data to process.

        Returns
        -------
        NDArray
            The processed data.
        """
        if self.At is not None:
            return self.At(x)
        else:
            return self.A(x)

    def absolute(self) -> "TransformFunctions":
        """Compute the absolute value of the operator. Raise an error, because not supported.

        Raises
        ------
        AttributeError
            Not supported operation.
        """
        raise AttributeError("Callable transform class does not support computing its absolute value.")


class ProjectorOperator(BaseTransform):
    """Base projector class that fixes the projection interface."""

    @property
    def vol_shape(self) -> NDArrayInt:
        """Expose the direct space shape as volume shape.

        Returns
        -------
        NDArray
            The volume shape.
        """
        return self.dir_shape

    @property
    def prj_shape(self) -> NDArrayInt:
        """Expose the adjoint space shape as projection shape.

        Returns
        -------
        NDArray
            The projection shape.
        """
        return self.adj_shape

    @vol_shape.setter
    def vol_shape(self, new_shape: Union[Sequence[int], NDArray]) -> None:
        self.dir_shape = np.array(new_shape, ndmin=1, dtype=int)

    @prj_shape.setter
    def prj_shape(self, new_shape: Union[Sequence[int], NDArray]) -> None:
        self.adj_shape = np.array(new_shape, ndmin=1, dtype=int)

    @abstractmethod
    def fp(self, x: NDArray) -> NDArray:
        """
        Define the interface of the forward-projection.

        Parameters
        ----------
        x : NDArray
            Input volume.

        Returns
        -------
        NDArray
            The projection data.
        """

    @abstractmethod
    def bp(self, x: NDArray) -> NDArray:
        """
        Define the interface for the back-projection.

        Parameters
        ----------
        x : NDArray
            Input projection data.

        Returns
        -------
        NDArray
            The back-projected volume.
        """

    def _op_direct(self, x: NDArray) -> NDArray:
        return self.fp(x)

    def _op_adjoint(self, x: NDArray) -> NDArray:
        return self.bp(x)

    def get_pre_weights(self) -> Union[NDArray, None]:
        """Compute the pre-weights of the projector geometry (notably for cone-beam geometries).

        Returns
        -------
        Union[NDArray, None]
            The computed detector weights
        """
        return None


class TransformIdentity(BaseTransform):
    """Identity operator."""

    def __init__(self, x_shape: Union[ArrayLike, NDArray]):
        """Identity operator.

        Parameters
        ----------
        x_shape : ArrayLike | NDArray
            Shape of the data.
        """
        self.dir_shape = np.array(x_shape, ndmin=1, dtype=int)
        self.adj_shape = np.array(x_shape, ndmin=1, dtype=int)
        super().__init__()

    def _op_direct(self, x: NDArray) -> NDArray:
        return x

    def _op_adjoint(self, x: NDArray) -> NDArray:
        return x


class TransformDiagonalScaling(BaseTransform):
    """Diagonal scaling operator."""

    scale: NDArray

    def __init__(self, x_shape: Union[ArrayLike, NDArray], scale: Union[ArrayLike, NDArray]):
        """Diagonal scaling operator.

        Parameters
        ----------
        x_shape : ArrayLike
            Shape of the data.
        scale : float or ArrayLike
            Operator diagonal.
        """
        self.scale = np.array(scale)
        self.dir_shape = np.array(x_shape, ndmin=1, dtype=int)
        self.adj_shape = np.array(x_shape, ndmin=1, dtype=int)
        super().__init__()

    def absolute(self) -> "TransformDiagonalScaling":
        """Return the projection operator using the absolute value of the projection coefficients.

        Returns
        -------
        TransformDiagonalScaling
            The absolute value operator
        """
        return TransformDiagonalScaling(self.dir_shape, np.abs(self.scale))

    def _op_direct(self, x: NDArray) -> NDArray:
        return self.scale * x

    def _op_adjoint(self, x: NDArray) -> NDArray:
        return self.scale * x


class TransformConvolution(BaseTransform):
    """
    Convolution operator.

    Parameters
    ----------
    x_shape : ArrayLike
        Shape of the direct space.
    kernel : ArrayLike
        The convolution kernel.
    pad_mode: str, optional
        The padding mode to use for the linear convolution. The default is "edge".
    is_symm : bool, optional
        Whether the operator is symmetric or not. The default is True.
    flip_adjoint : bool, optional
        Whether the adjoint kernel should be flipped. The default is False.
        This is useful when the kernel is not symmetric.
    """

    kernel: NDArray
    pad_mode: str
    is_symm: bool
    flip_adjoint: bool

    def __init__(
        self, x_shape: ArrayLike, kernel: ArrayLike, pad_mode: str = "edge", is_symm: bool = True, flip_adjoint: bool = False
    ):
        self.dir_shape = np.array(x_shape, ndmin=1, dtype=int)
        self.adj_shape = np.array(x_shape, ndmin=1, dtype=int)

        self.kernel = np.array(kernel, ndmin=len(self.dir_shape))

        self.pad_mode = pad_mode.lower()
        self.is_symm = is_symm
        self.flip_adjoint = flip_adjoint

        super().__init__()

    def absolute(self) -> "TransformConvolution":
        """
        Return the convolution operator using the absolute value of the kernel coefficients.

        Returns
        -------
        TransformConvolution
            The absolute value of the convolution operator.
        """
        return TransformConvolution(self.dir_shape, np.abs(self.kernel))

    def _pad_valid(self, x: NDArray) -> tuple[NDArray, NDArray]:
        pad_width = (np.array(self.kernel.shape) - 1) // 2
        return np.pad(x, pad_width=pad_width[:, None], mode=self.pad_mode), pad_width  # type: ignore

    def _crop_valid(self, x: NDArray, pad_width: NDArray) -> NDArray:
        slices = [slice(pw if pw else None, -pw if pw else None) for pw in pad_width]
        return x[tuple(slices)]

    def _op_direct(self, x: NDArray) -> NDArray:
        x, pw = self._pad_valid(x)
        x = spsig.convolve(x, self.kernel, mode="same")
        return self._crop_valid(x, pw)

    def _op_adjoint(self, x: NDArray) -> NDArray:
        if self.is_symm:
            x, pw = self._pad_valid(x)
            if self.flip_adjoint:
                adj_kernel = np.flip(self.kernel)
            else:
                adj_kernel = self.kernel
            x = spsig.convolve(x, adj_kernel, mode="same")
            return self._crop_valid(x, pw)
        else:
            return x


class BaseWaveletTransform(BaseTransform, ABC):
    """Base Wavelet transform."""

    axes: NDArrayInt
    wavelet: str
    labels: list[str]

    wlet_dec_filter_mult: NDArray
    wlet_rec_filter_mult: NDArray

    def _initialize_filter_bank(self) -> None:
        num_axes = len(self.axes)
        self.labels = [bin(x)[2:].zfill(num_axes).replace("0", "a").replace("1", "d") for x in range(1, 2**num_axes)]

        self.w = pywt.Wavelet(self.wavelet)  # type: ignore
        filt_bank_l1norm = np.linalg.norm(self.w.filter_bank, ord=1, axis=-1)
        self.wlet_dec_filter_mult = np.array(
            [(filt_bank_l1norm[0] ** lab.count("a")) * (filt_bank_l1norm[1] ** lab.count("d")) for lab in self.labels]
        )
        self.wlet_rec_filter_mult = np.array(
            [(filt_bank_l1norm[2] ** lab.count("a")) * (filt_bank_l1norm[3] ** lab.count("d")) for lab in self.labels]
        )


class TransformDecimatedWavelet(BaseWaveletTransform):
    """Decimated wavelet Transform operator."""

    def __init__(
        self,
        x_shape: Union[ArrayLike, NDArray],
        wavelet: str,
        level: int,
        axes: Optional[ArrayLike] = None,
        pad_on_demand: str = "edge",
    ):
        """
        Decimated wavelet Transform operator.

        Parameters
        ----------
        x_shape : ArrayLike
            Shape of the data to be wavelet transformed.
        wavelet : str
            Wavelet type.
        level : int
            Number of wavelet decomposition levels.
        axes : int or tuple of int, optional
            Axes along which to do the transform. Defaults to None.
        pad_on_demand : str, optional
            Padding type to fit the `2 ** level` shape requirements. Defaults to 'edge'.
            Options are all the `numpy.pad` padding modes.

        Raises
        ------
        ValueError
            If the pywavelets package is not available or its version is not adequate.
        """
        x_shape = np.array(x_shape, ndmin=1, dtype=int)

        if not has_pywt:
            raise ValueError("Cannot use Wavelet transform because pywavelets is not installed.")

        self.wavelet = wavelet
        self.level = level

        if axes is None:
            axes = np.arange(-len(x_shape), 0, dtype=int)
        self.axes = np.array(axes, ndmin=1, dtype=int)

        self.pad_on_demand = pad_on_demand

        self._initialize_filter_bank()

        num_axes = len(self.axes)

        self.dir_shape = x_shape

        self.sub_band_shapes = pywt.wavedecn_shapes(
            self.dir_shape, self.wavelet, mode=self.pad_on_demand, level=self.level, axes=self.axes
        )
        self.adj_shape = self.dir_shape.copy()
        for ax in self.axes:
            self.adj_shape[ax] = self.sub_band_shapes[0][ax] + np.sum(
                [self.sub_band_shapes[x]["d" * num_axes][ax] for x in range(1, self.level + 1)]
            )
        self.slicing_info = None

        super().__init__()

    def direct_dwt(self, x: NDArray) -> list:
        """
        Perform the direct wavelet transform.

        Parameters
        ----------
        x : NDArray
            Data to transform.

        Returns
        -------
        list
            Transformed data.
        """
        return pywt.wavedecn(x, wavelet=self.wavelet, axes=self.axes, mode=self.pad_on_demand, level=self.level)

    def inverse_dwt(self, y: list) -> NDArray:
        """
        Perform the inverse wavelet transform.

        Parameters
        ----------
        y : list
            Data to anti-transform.

        Returns
        -------
        NDArray
            Anti-transformed data.
        """
        rec = pywt.waverecn(y, wavelet=self.wavelet, axes=self.axes, mode=self.pad_on_demand)
        if not np.all(rec.shape == self.dir_shape):
            slices = [slice(0, s) for s in self.dir_shape]
            rec = rec[tuple(slices)]
        return rec

    def _op_direct(self, x: NDArray) -> NDArray:
        c = self.direct_dwt(x)
        y, self.slicing_info = pywt.coeffs_to_array(c, axes=self.axes)
        return y

    def _op_adjoint(self, x: NDArray) -> NDArray:
        if self.slicing_info is None:
            _ = self._op_direct(np.zeros(self.dir_shape))

        c = pywt.array_to_coeffs(x, self.slicing_info)
        return self.inverse_dwt(c)


class TransformStationaryWavelet(BaseWaveletTransform):
    """Stationary wavelet Transform operator."""

    def __init__(
        self,
        x_shape: ArrayLike,
        wavelet: str,
        level: int,
        axes: Optional[ArrayLike] = None,
        pad_on_demand: str = "edge",
        normalized: bool = True,
    ):
        """Stationary wavelet Transform operator.

        Parameters
        ----------
        x_shape : ArrayLike
            The shape of the data to be wavelet transformed.
        wavelet : str
            The type of wavelet to use.
        level : int
            Number of wavelet decomposition levels.
        axes : int or tuple of int, optional
            Axes along which to perform the transform. Default is None.
        pad_on_demand : str, optional
            The padding type to fit the `2 ** level` shape requirements.
            Default is 'constant'. Options are all the `numpy.pad` padding modes.
        normalized : bool, optional
            Whether to use a normalized transform. Default is True.

        Raises
        ------
        ValueError
            If the pywavelets package is not available or its version is not adequate.
        """
        x_shape = np.array(x_shape, ndmin=1, dtype=int)

        if not has_pywt:
            raise ValueError("Cannot use Wavelet transform because pywavelets is not installed.")
        if not use_swtn:
            raise ValueError("Cannot use Wavelet transform because pywavelets is too old (<1.0.2).")

        self.wavelet = wavelet
        self.level = level
        self.normalized = normalized

        if axes is None:
            axes = np.arange(-len(x_shape), 0, dtype=int)
        self.axes = np.array(axes, ndmin=1, dtype=int)

        self.pad_on_demand = pad_on_demand

        self._initialize_filter_bank()

        self.dir_shape = x_shape
        if self.pad_on_demand is not None:
            alignment = 2**self.level
            x_axes = np.array(self.dir_shape)[np.array(self.axes)]
            self.pad_axes = (alignment - x_axes % alignment) % alignment

            adj_x_shape = cp.deepcopy(self.dir_shape)
            adj_x_shape[np.array(self.axes)] += self.pad_axes
        else:
            adj_x_shape = self.dir_shape
        self.adj_shape = np.array((self.level * (2 ** len(self.axes) - 1) + 1, *adj_x_shape))

        super().__init__()

    def direct_swt(self, x: NDArray) -> list:
        """
        Perform the direct wavelet transform.

        Parameters
        ----------
        x : NDArray
            Data to transform.

        Returns
        -------
        list
            Transformed data.
        """
        if self.pad_on_demand is not None and np.any(self.pad_axes):
            for ax in np.nonzero(self.pad_axes)[0]:
                pad_l = np.ceil(self.pad_axes[ax] / 2).astype(int)
                pad_h = np.floor(self.pad_axes[ax] / 2).astype(int)
                pad_width = [(0, 0)] * len(x.shape)
                pad_width[self.axes[ax]] = (pad_l, pad_h)
                x = np.pad(x, pad_width, mode=self.pad_on_demand)
        return pywt.swtn(x, wavelet=self.wavelet, axes=self.axes, norm=self.normalized, level=self.level, trim_approx=True)

    def inverse_swt(self, y: list) -> NDArray:
        """
        Perform the inverse wavelet transform.

        Parameters
        ----------
        y : list
            Data to anti-transform.

        Returns
        -------
        NDArray
            Anti-transformed data.
        """
        x = pywt.iswtn(y, wavelet=self.wavelet, axes=self.axes, norm=self.normalized)
        if self.pad_on_demand is not None and np.any(self.pad_axes):
            for ax in np.nonzero(self.pad_axes)[0]:
                pad_l = np.ceil(self.pad_axes[ax] / 2).astype(int)
                pad_h = np.floor(self.pad_axes[ax] / 2).astype(int)
                slices = [slice(None)] * len(x.shape)
                slices[self.axes[ax]] = slice(pad_l, x.shape[self.axes[ax]] - pad_h, 1)
                x = x[tuple(slices)]
        return x

    def _op_direct(self, x: NDArray) -> NDArray:
        y = self.direct_swt(x)
        y = [y[0]] + [y[lvl][x] for lvl in range(1, self.level + 1) for x in self.labels]
        return np.array(y)

    def _op_adjoint(self, x: NDArray) -> NDArray:
        def get_lvl_pos(lvl):
            return (lvl - 1) * (2 ** len(self.axes) - 1) + 1

        y = [x[0]] + [
            {k: x[ii_lbl + get_lvl_pos(lvl), ...] for ii_lbl, k in enumerate(self.labels)} for lvl in range(1, self.level + 1)
        ]
        return self.inverse_swt(y)


class TransformGradient(BaseTransform):
    """
    Gradient operator.

    Parameters
    ----------
    x_shape : ArrayLike
        Shape of the data to be transformed.
    axes : Optional[ArrayLike], optional
        Axes along which to do the gradient. The default is None.
    pad_mode : str, optional
        Padding mode of the gradient. The default is "edge".
    """

    def __init__(self, x_shape: ArrayLike, axes: Optional[ArrayLike] = None, pad_mode: str = "edge"):
        x_shape = np.array(x_shape, ndmin=1, dtype=int)

        if axes is None:
            axes = np.arange(-len(x_shape), 0, dtype=int)
        self.axes = np.array(axes, ndmin=1, dtype=int)
        self.ndims = len(x_shape)

        self.pad_mode = pad_mode.lower()

        self.dir_shape = x_shape
        self.adj_shape = np.array((len(self.axes), *self.dir_shape), ndmin=1, dtype=int)

        super().__init__()

    def gradient(self, x: NDArray) -> NDArray:
        """
        Compute the gradient.

        Parameters
        ----------
        x : NDArray
            Input data.

        Returns
        -------
        NDArray
            Gradient of data.
        """
        d = [np.array([])] * len(self.axes)
        for ii, ax in enumerate(self.axes):
            padding = np.zeros((self.ndims, 2), dtype=int)
            padding[ax, 1] = 1
            temp_x = np.pad(x, padding, mode=self.pad_mode)  # type: ignore
            d[ii] = np.diff(temp_x, n=1, axis=ax)
        return np.stack(d, axis=0)

    def divergence(self, x: NDArray) -> NDArray:
        """
        Compute the divergence - transpose of gradient.

        Parameters
        ----------
        x : NDArray
            Input data.

        Returns
        -------
        NDArray
            Divergence of data.
        """
        d = [np.array([])] * len(self.axes)
        for ii, ax in enumerate(self.axes):
            padding = np.zeros((self.ndims, 2), dtype=int)
            padding[ax, 0] = 1
            temp_x = np.pad(x[ii, ...], padding, mode=self.pad_mode)  # type: ignore
            d[ii] = np.diff(temp_x, n=1, axis=ax)
        return np.sum(np.stack(d, axis=0), axis=0)

    def _op_direct(self, x: NDArray) -> NDArray:
        return self.gradient(x)

    def _op_adjoint(self, x: NDArray) -> NDArray:
        return -self.divergence(x)


class TransformFourier(BaseTransform):
    """Fourier transform operator."""

    def __init__(self, x_shape: ArrayLike, axes: Optional[ArrayLike] = None):
        """
        Fourier transform.

        Parameters
        ----------
        x_shape : ArrayLike
            Shape of the data to be Fourier transformed.
        axes : Optional[ArrayLike], optional
            Axes along which to do the Fourier transform.

        Returns
        -------
        None
        """
        x_shape = np.array(x_shape, ndmin=1, dtype=int)

        if axes is None:
            axes = np.arange(-len(x_shape), 0, dtype=int)
        self.axes = np.array(axes, ndmin=1, dtype=int)
        self.ndims = len(x_shape)

        self.dir_shape = x_shape
        self.adj_shape = np.array((2, *self.dir_shape), ndmin=1, dtype=int)

        super().__init__()

    def fft(self, x: NDArray) -> NDArray:
        """
        Compute the fft.

        Parameters
        ----------
        x : NDArray
            Input data.

        Returns
        -------
        NDArray
            FFT of data.
        """
        d = np.empty(self.adj_shape, dtype=x.dtype)
        x_f = np.fft.fftn(x, axes=tuple(self.axes), norm="ortho")
        d[0, ...] = x_f.real
        d[1, ...] = x_f.imag
        return d

    def ifft(self, x: NDArray) -> NDArray:
        """Compute the inverse of the fft.

        Parameters
        ----------
        x : NDArray
            Input data.

        Returns
        -------
        NDArray
            iFFT of data.
        """
        d = x[0, ...] + 1j * x[1, ...]
        return np.fft.ifftn(d, axes=tuple(self.axes), norm="ortho").real

    def _op_direct(self, x: NDArray) -> NDArray:
        return self.fft(x)

    def _op_adjoint(self, x: NDArray) -> NDArray:
        return self.ifft(x)


class TransformLaplacian(BaseTransform):
    """
    Laplacian transform operator.

    Parameters
    ----------
    x_shape : ArrayLike
        Shape of the data to be transformed.
    axes : ArrayLike, optional
        Axes along which to do the Laplacian. The default is None.
    pad_mode : str, optional
        Padding mode of the Laplacian. The default is "edge".
    """

    def __init__(self, x_shape: ArrayLike, axes: Optional[ArrayLike] = None, pad_mode: str = "edge"):
        x_shape = np.array(x_shape, ndmin=1, dtype=int)

        if axes is None:
            axes = np.arange(-len(x_shape), 0, dtype=int)
        self.axes = np.array(axes, ndmin=1, dtype=int)
        self.ndims = len(x_shape)

        self.pad_mode = pad_mode.lower()

        self.dir_shape = x_shape
        self.adj_shape = x_shape

        super().__init__()

    def laplacian(self, x: NDArray) -> NDArray:
        """Compute the laplacian.

        Parameters
        ----------
        x : NDArray
            Input data.

        Returns
        -------
        NDArray
            Laplacian of the input data.
        """
        d = [np.array([])] * len(self.axes)
        for ii, ax in enumerate(self.axes):
            padding = np.zeros((self.ndims, 2), dtype=int)
            padding[ax, :] = 1
            temp_x = np.pad(x, padding, mode=self.pad_mode)  # type: ignore
            d[ii] = np.diff(temp_x, n=2, axis=ax)
        return np.sum(d, axis=0)

    def _op_direct(self, x: NDArray) -> NDArray:
        return self.laplacian(x)

    def _op_adjoint(self, x: NDArray) -> NDArray:
        return self.laplacian(x)


class TransformSVD(BaseTransform):
    """Singular value decomposition based decomposition operator."""

    U: Optional[NDArray]
    Vt: Optional[NDArray]

    rescale: bool

    def __init__(
        self,
        x_shape: ArrayLike,
        axes_rows: Union[Sequence[int], NDArray] = (0,),
        axes_cols: Union[Sequence[int], NDArray] = (-1,),
        rescale: bool = False,
    ):
        """
        Singular value decomposition operator.

        The SVD decomposition will be done over the flattened rows vs flatted cols.
        This means that the channels should always be the rows (expected to be only
        one dimension, usually), while the volume dimensions should always be the
        columns (expected to be the last two or three dimensions).

        Parameters
        ----------
        x_shape : numpy.array_like
            Shape of the data to be wavelet transformed.
        axes_rows : tuple of int, optional
            Axes expanded in rows of the SVD. Defaults to (0, ).
        axes_cols : tuple of int, optional
            Axes expanded in cols of the SVD. Defaults to (-1, ).

        Raises
        ------
        IndexError
            In case the axes are outside the range.
        """
        self.dir_shape = np.array(x_shape, ndmin=1, dtype=int)

        self.axes_rows = np.array(axes_rows, ndmin=1) % len(self.dir_shape)
        self.axes_cols = np.array(axes_cols, ndmin=1) % len(self.dir_shape)

        # Dimensions to decompose
        self.append_dims = np.concatenate((self.axes_rows, self.axes_cols)).astype(int)
        # Dimensions to NOT decompose
        temp_dims = np.arange(len(self.dir_shape), dtype=int)
        self.invariant_dims = np.delete(temp_dims, self.append_dims)
        self.invariant_dims_shape = self.dir_shape[self.invariant_dims]

        # Transpose operation to prepare data for decomposition
        self.fwd_transpose = np.concatenate((self.invariant_dims, self.append_dims))
        # Transpose operation to recover data after re-composition
        self.bwd_transpose = np.argsort(self.fwd_transpose)

        self.axes_rows_shape = self.dir_shape[self.axes_rows]
        self.axes_cols_shape = self.dir_shape[self.axes_cols]
        self.axes_rows_size = (np.prod(self.axes_rows_shape),)
        self.axes_cols_size = (np.prod(self.axes_cols_shape),)

        # Reshape operation to prepare data for decomposition
        self.fwd_shape = np.concatenate((self.invariant_dims_shape, self.axes_rows_size, self.axes_cols_size))
        # Reshape operation to recover data after re-composition
        self.bwd_shape = np.concatenate((self.invariant_dims_shape, self.axes_rows_shape, self.axes_cols_shape))

        self.adj_shape = np.concatenate((self.invariant_dims_shape, np.fmin(self.axes_rows_size, self.axes_cols_size)))

        self.U = None
        self.Vt = None

        self.rescale = rescale

        super().__init__()

    def direct_svd(self, x: NDArray):
        """
        Performs the SVD decomposition.

        Parameters
        ----------
        x : NDArray
            Data to transform.

        Returns
        -------
        tuple(U, s, Vt)
            Transformed data.
        """
        return np.linalg.svd(x, full_matrices=False, compute_uv=True)

    def inverse_svd(self, U: NDArray, s: NDArray, Vt: NDArray) -> NDArray:
        """
        Performs the inverse SVD decomposition.

        Parameters
        ----------
        U : NDArray
            Rows of the SVD decomposition.
        s : NDArray
            Singular values.
        Vt : NDArray
            Columns of the SVD decomposition.

        Returns
        -------
        NDArray
            Anti-transformed data.
        """
        return np.matmul(U, s[..., None] * Vt)

    def _op_direct(self, x: NDArray) -> NDArray:
        x = np.transpose(x, self.fwd_transpose)
        x = np.reshape(x, self.fwd_shape)

        (self.U, s, self.Vt) = self.direct_svd(x)

        if self.rescale:
            s /= np.sqrt(self.Vt.shape[-1] * self.U.shape[-2])
            self.Vt *= np.sqrt(self.Vt.shape[-1])
            self.U *= np.sqrt(self.U.shape[-2])

        return s

    def _op_adjoint(self, x: NDArray) -> NDArray:
        if self.U is None or self.Vt is None:
            raise ValueError("Operator not initialized!")

        y = self.inverse_svd(self.U, x, self.Vt)

        y = np.reshape(y, self.bwd_shape)
        return np.transpose(y, self.bwd_transpose)


if __name__ == "__main__":
    test_vol = np.zeros((10, 10), dtype=np.float32)

    H = TransformStationaryWavelet(test_vol.shape, "db1", 2)
    Htw = H.T.explicit()
    Hw = H.explicit()

    D = TransformGradient(test_vol.shape)
    Dg = D.explicit()



"""
Noise handling routines.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import numpy as np

from typing import Optional
from numpy.typing import NDArray


eps = np.finfo(np.float32).eps


def compute_variance_poisson(
    Is: NDArray, I0: Optional[NDArray] = None, var_I0: Optional[NDArray] = None, normalized: bool = True
) -> NDArray:
    """
    Compute the variance of a signal subject to Poisson noise, against a reference intensity.

    The reference intensity can also be subject to Poisson and Gaussian noise.
    If the variance of the reference intensity is not passed, it will be assumed to be Poisson.

    Parameters
    ----------
    Is : NDArray
        The signal intensity.
    I0 : Optional[NDArray], optional
        The reference intensity. The default is None.
    var_I0 : Optional[NDArray], optional
        The variance of the reference intensity. The default is None.
        If not given, it will be assumed to be equal to I0.
    normalized : bool, optional
        Whether to renormalize by the mean of the reference intensity.

    Returns
    -------
    NDArray
        The computed variance.
    """
    var_Is = np.abs(Is)
    Is = np.fmax(Is, eps)

    if I0 is not None:
        if var_I0 is None:
            var_I0 = np.abs(I0)
        I0 = np.fmax(I0, eps)

        Is2 = Is**2
        I02 = I0**2
        variance = (Is2 / I02) * (var_Is / Is2 + var_I0 / I02)
        if normalized:
            variance *= np.mean(I0)
        return variance
    else:
        return var_Is


def compute_variance_transmission(
    Is: NDArray, I0: NDArray, var_I0: Optional[NDArray] = None, normalized: bool = True
) -> NDArray:
    """
    Compute the variance of a linearized attenuation (transmission) signal, against a reference intensity.

    Parameters
    ----------
    Is : NDArray
        The transmitted signal.
    I0 : NDArray
        The reference intensity.
    var_I0 : Optional[NDArray], optional
        The variance of the reference intensity. The default is None.
        If not given, it will be assumed to be equal to I0.
    normalized : bool, optional
        Whether to renormalize by the mean of the reference intensity.

    Returns
    -------
    NDArray
        The computed variance.
    """
    var_Is = np.abs(Is)
    Is = np.fmax(Is, eps)

    if var_I0 is None:
        var_I0 = np.abs(I0)
    I0 = np.fmax(I0, eps)

    Is2 = Is**2
    I02 = I0**2
    variance = (Is / I0) * (var_Is / Is2 + var_I0 / I02)
    if normalized:
        variance *= np.mean(I0)
    return variance


def compute_variance_weight(
    variance: NDArray,
    *,
    percentile: float = 0.001,
    mask: Optional[NDArray] = None,
    normalized: bool = False,
    use_std: bool = False,
    semilog: bool = False
) -> NDArray:
    """
    Compute the weight associated to the given variance, in a weighted least-squares context.

    Parameters
    ----------
    variance : NDArray
        The variance of the signal.
    percentile : float
        Minimum percentile to discard. The default is 0.001 (0.1%)
    mask : NDArray | None, optional
        Mask of valid values. The default is None.
    normalized : bool, optional
        Scale the largest weight to 1. The default is False.
    use_std : bool, optional
        Use the standard deviation instead of the variance.
    semilog : bool, optional
        Scale the variance over a logarithmic curve. It can be beneficial with
        high dynamic range data. The default is False.

    Returns
    -------
    NDArray
        The computed weights.
    """
    variance = np.abs(variance)

    vals = variance.flatten()
    if mask is not None:
        vals = vals[mask.flatten()]

    sorted_variances = np.sort(vals)
    percentiles_variances = np.cumsum(sorted_variances) / np.sum(sorted_variances)
    ind_threshold = np.fmin(np.fmax(0, np.sum(percentiles_variances < percentile)), percentiles_variances.size)
    min_val = np.fmax(sorted_variances[ind_threshold], eps)

    min_nonzero_variance = np.fmax(np.min(vals[vals > 0]), min_val)
    inv_weight = np.fmax(variance, min_nonzero_variance)

    if normalized:
        inv_weight /= min_nonzero_variance
    if use_std:
        inv_weight = np.sqrt(inv_weight)
    if semilog:
        inv_weight = np.log(inv_weight + float(normalized is False)) + 1

    weight = 1 / inv_weight

    if mask is not None:
        weight *= mask.astype(weight.dtype)

    return weight



# -*- coding: utf-8 -*-

"""Package for data alignment."""

__author__ = """Nicola VIGANÒ"""
__email__ = "N.R.Vigano@cwi.nl"

from . import fitting  # noqa: F401, F402
from . import centering  # noqa: F401, F402
from . import shifts  # noqa: F401, F402

RecenterVolume = centering.RecenterVolume
DetectorShiftsPRE = shifts.DetectorShiftsPRE
DetectorShiftsXC = shifts.DetectorShiftsXC



"""
Post-processing routines.

Created on Tue Mar 24 15:25:14 2020

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

from collections.abc import Sequence
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import scipy as sp
from matplotlib.axes._axes import Axes
from matplotlib.figure import Figure
from numpy.typing import ArrayLike, NDArray
from scipy.optimize import minimize

from tqdm.auto import tqdm

from corrct.operators import BaseTransform, TransformIdentity
from corrct.processing.misc import azimuthal_integration, circular_mask, lines_intersection


eps = np.finfo(np.float32).eps


def com(vol: NDArray, axes: Optional[ArrayLike] = None) -> NDArray:
    """
    Compute center-of-mass for given volume.

    Parameters
    ----------
    vol : NDArray
        The input volume.
    axes : ArrayLike, optional
        Axes on which to compute center-of-mass. The default is None.

    Returns
    -------
    NDArray
        The center-of-mass position.
    """
    if axes is None:
        axes = np.arange(len(vol.shape))
    else:
        axes = np.array(axes, ndmin=1)

    coords = [np.linspace(-(s - 1) / 2, (s - 1) / 2, s) for s in np.array(vol.shape)[list(axes)]]

    num_dims = len(vol.shape)
    center_of_mass = np.empty((len(axes),))
    for ii, a in enumerate(axes):
        sum_axes = np.array(np.delete(np.arange(num_dims), a), ndmin=1, dtype=int)
        line = np.abs(vol).sum(axis=tuple(sum_axes))
        center_of_mass[ii] = line.dot(coords[ii]) / line.sum()

    return center_of_mass


def power_spectrum(
    img: NDArray,
    axes: Optional[Sequence[int]] = None,
    smooth: Optional[int] = 5,
    taper_ratio: Optional[float] = 0.05,
    power: int = 2,
) -> NDArray:
    """
    Compute the power spectrum of a n-dimensional signal.

    Parameters
    ----------
    img : NDArray
        The n-dimensional signal.
    axes : Optional[Sequence[int]], optional
        The axes over which we want to compute the power spectrum, by default None
    smooth : Optional[int], optional
        The smoothing kernel size, by default 5
    taper_ratio : Optional[float], optional
        Whether to taper the signal at the edges (for truncated signals), by default 0.05
    power : int, optional
        The exponent to use, by default 2

    Returns
    -------
    NDArray
        The power spectrum
    """
    img_shape = np.array(img.shape)

    if axes is None:
        axes = list(np.arange(-len(img_shape), 0))

    axes_shape = img_shape[list(axes)]
    cut_off = np.min(axes_shape) // 2

    if taper_ratio is not None:
        taper_size = float(taper_ratio * np.mean(axes_shape))
        vol_mask = circular_mask(img_shape, coords_ball=axes, radius_offset=-taper_size, taper_func="cos")
        img = img * vol_mask

    img_f = np.fft.fftn(img, axes=axes)

    f1 = np.abs(img_f) ** power
    f1_int = azimuthal_integration(f1, axes=axes, domain="fourier")

    rings_size = azimuthal_integration(np.ones_like(img), axes=axes, domain="fourier")
    ps = f1_int / rings_size
    dc_val = np.sqrt(np.min(axes_shape) ** len(axes_shape)) ** power
    ps /= dc_val

    if smooth is not None and smooth > 1:
        win = sp.signal.windows.hann(smooth)
        win /= np.sum(win)
        win = win.reshape([*[1] * (ps.ndim - 1), -1])
        ps = sp.ndimage.convolve(ps, win, mode="nearest")

    return ps[..., :cut_off]


def compute_frc(
    img1: NDArray,
    img2: Optional[NDArray],
    snrt: float = 0.2071,
    axes: Optional[Sequence[int]] = None,
    smooth: Optional[int] = 5,
    taper_ratio: Optional[float] = 0.05,
    supersampling: int = 1,
    theo_threshold: bool = True,
) -> tuple[NDArray, NDArray]:
    """
    Compute the FRC/FSC (Fourier ring/shell correlation) between two images / volumes.

    Please refer to the following article for more information:
        M. van Heel and M. Schatz, “Fourier shell correlation threshold criteria,”
        J. Struct. Biol., vol. 151, no. 3, pp. 250–262, Sep. 2005.

    Parameters
    ----------
    img1 : NDArray
        First image / volume.
    img2 : NDArray
        Second image / volume.
    snrt : float, optional
        SNR to be used for generating the threshold curve for resolution definition.
        The SNR value of 0.4142 corresponds to the hald-bit curve for a full dataset.
        When splitting datasets in two sub-datasets, that value needs to be halved.
        The default is 0.2071, which corresponds to the half-bit threashold for half dataset.
    axes : Sequence[int], optional
        The axes over which we want to compute the FRC/FSC.
        If None, all axes will be used The default is None.
    smooth : Optional[int], optional
        Size of the Hann smoothing window. The default is 5.
    taper_ratio : Optional[float], optional
        Ratio of the edge pixels to be tapered off. This is necessary when working
        with truncated volumes / local tomography, to avoid truncation artifacts.
        The default is 0.05.
    supersampling : int, optional
        Supersampling factor of the images.
        Larger values increase the high-frequency range of the FRC/FSC function.
        The default is 1, which corresponds to the Nyquist frequency.

    Raises
    ------
    ValueError
        Error returned when not passing images of the same shape.

    Returns
    -------
    NDArray
        The computed FRC/FSC.
    NDArray
        The threshold curve corresponding to the given threshod SNR.
    """
    img1_shape = np.array(img1.shape)

    if axes is None:
        axes = list(np.arange(-len(img1_shape), 0))

    if img2 is None:
        if np.any(img1_shape[axes] % 2 == 1):
            raise ValueError(f"Image shape {img1_shape} along the chosen axes {axes} needs to be even.")
        raise NotImplementedError("Self FRC not implemented, yet.")
    else:
        img2_shape = np.array(img2.shape)
        if len(img1_shape) != len(img2_shape) or np.any(img1_shape != img2_shape):
            raise ValueError(
                f"Image #1 size {img1_shape} and image #2 size {img2_shape} are different, while they should be equal."
            )
    if img1.dtype != img2.dtype:
        print(f"WARNING: The two images have different dtype: img1 {img1.dtype}, img2 {img2.dtype}. Forcing the first.")
        img2 = img2.astype(img1.dtype)
    dtype = img1.dtype

    if supersampling > 1:
        # Bodge to make interpolation work with recent scipy: because the cython implementation does not compile for float32
        dtype = float
        img1 = img1.astype(dtype)
        img2 = img2.astype(dtype)

        base_grid = [np.linspace(-(d - 1) / 2, (d - 1) / 2, d, dtype=dtype) for d in img1_shape]

        interp_grid = [np.linspace(-(d - 1) / 2, (d - 1) / 2, d, dtype=dtype) for d in img1_shape]
        for a in axes:
            d = img1_shape[a] * 2
            interp_grid[a] = np.linspace(-(d - 1) / 4, (d - 1) / 4, d, dtype=dtype)
        interp_grid = np.meshgrid(*interp_grid, indexing="ij")
        interp_grid = np.transpose(interp_grid, [*range(1, len(img1_shape) + 1), 0])

        img1 = sp.interpolate.interpn(base_grid, img1, interp_grid, bounds_error=False, fill_value=None)
        img2 = sp.interpolate.interpn(base_grid, img2, interp_grid, bounds_error=False, fill_value=None)

        img1_shape = np.array(img1.shape)

    axes_shape = img1_shape[list(axes)]
    cut_off = np.min(axes_shape) // 2

    if taper_ratio is not None:
        taper_size = float(taper_ratio * np.mean(axes_shape))
        vol_mask = circular_mask(img1_shape, coords_ball=axes, radius_offset=-taper_size, taper_func="cos")
        img1 = img1 * vol_mask
        img2 = img2 * vol_mask

    img1_f = np.fft.fftn(img1, axes=axes)
    img2_f = np.fft.fftn(img2, axes=axes)

    fc = img1_f * np.conj(img2_f)
    f1 = np.abs(img1_f) ** 2
    f2 = np.abs(img2_f) ** 2

    fc_r_int = azimuthal_integration(fc.real, axes=axes, domain="fourier")
    fc_i_int = azimuthal_integration(fc.imag, axes=axes, domain="fourier")
    fc_int = np.sqrt((fc_r_int**2) + (fc_i_int**2))
    f1_int = azimuthal_integration(f1, axes=axes, domain="fourier")
    f2_int = azimuthal_integration(f2, axes=axes, domain="fourier")

    f1s_f2s = f1_int * f2_int
    f1s_f2s = f1s_f2s + (f1s_f2s == 0)
    f1s_f2s = np.sqrt(f1s_f2s)

    frc = fc_int / f1s_f2s

    if theo_threshold:
        # The number of pixels in a ring is given by the surface.
        # We compute the n-dimensional hyper-sphere surface, where n is given by the number of axes.
        n = len(axes)
        num_surf = 2 * np.pi ** (n / 2)
        den_surf = sp.special.gamma(n / 2)
        rings_size = np.concatenate(((1.0,), num_surf / den_surf * np.arange(1, len(frc)) ** (n - 1)))
    else:
        rings_size = azimuthal_integration(np.ones_like(img1), axes=axes, domain="fourier")

    t_num = snrt + (2 * np.sqrt(snrt) + 1) / np.sqrt(rings_size)
    t_den = snrt + 1 + 2 * np.sqrt(snrt) / np.sqrt(rings_size)
    t_hb = t_num / t_den

    if smooth is not None and smooth > 1:
        win = sp.signal.windows.hann(smooth)
        win /= np.sum(win)
        win = win.reshape([*[1] * (frc.ndim - 1), -1])
        frc = sp.ndimage.convolve(frc, win, mode="nearest")

    return frc[..., :cut_off], t_hb[..., :cut_off]


def estimate_resolution(frc: NDArray, t_hb: NDArray) -> Optional[tuple[float, float]]:
    """Estimate the resolution or bandwidth, given an FRC and a threshold curve.

    Parameters
    ----------
    frc : NDArray
        The FRC curve
    t_hb : NDArray
        The threshold curve

    Returns
    -------
    tuple[float, float] | None
        The resolution or bandwidth, if a crossing point was found. Otherwise None.
    """
    if t_hb.ndim > 1:
        reduce_axes = tuple(np.arange(t_hb.ndim - 1))
        frc = frc.mean(axis=reduce_axes)
        t_hb = t_hb.mean(axis=reduce_axes)
    return lines_intersection(frc, t_hb, x_lims=(1, None))


def plot_frcs(
    volume_pairs: Sequence[tuple[NDArray, NDArray]],
    labels: Sequence[str],
    title: Optional[str] = None,
    smooth: Optional[int] = 5,
    snrt: float = 0.2071,
    axes: Optional[Sequence[int]] = None,
    supersampling: int = 1,
    verbose: bool = False,
) -> tuple[Figure, Axes]:
    """Compute and plot the FSCs / FRCs of some volumes.

    Parameters
    ----------
    volume_pairs : Sequence[Tuple[NDArray, NDArray]]
        A list of pairs of volumes to compute the FRCs on.
    labels : Sequence[str]
        The labels associated with each pair.
    title : Optional[str], optional
        The axes title, by default None.
    smooth : Optional[int], optional
        The size of the smoothing window for the computed curves, by default 5.
    snrt : float, optional
        The SNR of the T curve, by default 0.2071 - as per half-dataset SNR.
    axes : Sequence[int] | None, optional
        The axes along which we want to compute the FRC. The unused axes will be
        averaged. The default is None.
    verbose : bool, optional
        Whether to display verbose output, by default False.
    """
    frcs = [np.array([])] * len(volume_pairs)
    xps: list[Optional[tuple[float, float]]] = [(0.0, 0.0)] * len(volume_pairs)

    for ii, pair in enumerate(tqdm(volume_pairs, desc="Computing FRCs", disable=not verbose)):
        frcs[ii], t_hb = compute_frc(pair[0], pair[1], snrt=snrt, smooth=smooth, axes=axes, supersampling=supersampling)
        xps[ii] = estimate_resolution(frcs[ii], t_hb)

    nyquist = len(frcs[0])
    xx = np.linspace(0, 1, nyquist)

    fig, axs = plt.subplots(1, 1, sharex=True, sharey=True)
    for f, l in zip(frcs, labels):
        axs.plot(xx, np.squeeze(f), label=l)
    axs.plot(xx, np.squeeze(t_hb), label="T 1/2 bit", linestyle="dashed")
    for ii, p in enumerate(xps):
        if p is not None:
            res = p[0] / (nyquist - 1)
            axs.stem(res, p[1], label=f"Resolution ({labels[ii]}): {res:.3}", linefmt=f"C{ii}-.", markerfmt=f"C{ii}o")
    axs.set_xlim(0, 1)
    axs.set_ylim(0, None)
    axs.legend(fontsize=12)
    axs.grid()
    axs.set_ylabel("Magnitude", fontdict=dict(fontsize=16))
    axs.set_xlabel("Spatial frequency / Nyquist", fontdict=dict(fontsize=16))
    if title is not None:
        axs.set_title(title)
    for tl in axs.get_xticklabels():
        tl.set_fontsize(13)
    for tl in axs.get_yticklabels():
        tl.set_fontsize(13)
    fig.tight_layout()

    plt.show(block=False)

    return fig, axs


def fit_scale_bias(img_data: NDArray, prj_data: NDArray, prj: Optional[BaseTransform] = None) -> tuple[float, float]:
    """Fit the scale and bias of an image, against its projection in a different space.

    Parameters
    ----------
    img_data : NDArray
        The image data
    prj_data : NDArray
        The projected data
    prj : BaseTransform | None, optional
        The projection operator. The default is None, which uses the identity (TransformIdentity)

    Returns
    -------
    tuple[float, float]
        The scale and bias
    """
    if prj is None:
        prj = TransformIdentity(img_data.shape)

    prj_x = prj(img_data)
    prj_1 = prj(np.ones_like(img_data))
    m_y_dot_prj_x = -float(np.sum(prj_data * prj_x))
    m_y_dot_prj_1 = -float(np.sum(prj_data * prj_1))
    prj_x_2 = float(np.sum(prj_x**2))
    prj_1_2 = float(np.sum(prj_1**2))
    prj_1_dot_prj_x = float(np.sum(prj_1 * prj_x))

    def obj_func(ab: NDArray) -> tuple[float, NDArray]:
        residual = prj(img_data * ab[0] + ab[1]) - prj_data
        grad_a = m_y_dot_prj_x + prj_x_2 * ab[0] + prj_1_dot_prj_x * ab[1]
        grad_b = m_y_dot_prj_1 + prj_1_2 * ab[1] + prj_1_dot_prj_x * ab[0]
        return float(np.linalg.norm(residual, ord=2) ** 2) / 2, np.array((grad_a, grad_b))

    opt_res = minimize(obj_func, [1.0, 0.0], jac=True)
    return float(opt_res.x[0]), float(opt_res.x[1])



#!/usr/bin/env python3
"""
Provide utility functions for testing.

Created on Thu Jun  4 12:28:21 2020

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import numpy as np
from . import projectors, processing
from . import physics

from collections.abc import Sequence
from typing import Optional
from typing import Union
from numpy.typing import DTypeLike
from numpy.typing import NDArray
import skimage.data as skd
import skimage.filters as skf
import skimage.transform as skt
import skimage.segmentation as sks
import skimage.morphology as skm


NDArrayFloat = NDArray[np.floating]


def roundup_to_pow2(x: Union[int, float, NDArrayFloat], p: int, dtype: DTypeLike = int) -> Union[int, float, NDArrayFloat]:
    """Round first argument to the power of 2 indicated by second argument.

    Parameters
    ----------
    x : int | float | NDArrayFloat
        Number to round up.
    p : int
        Power of 2.
    dtype : DTypeLike, optional
        data type of the output. The default is int.

    Returns
    -------
    int | float | NDArrayFloat
        Rounding up of input.
    """
    return np.ceil(np.array(x) / (2**p)).astype(dtype) * (2**p)


def download_phantom():
    """Download the phantom generation module from Nicolas Barbey's repository on github."""
    phantom_url = "https://raw.githubusercontent.com/nbarbey/TomograPy/master/tomograpy/phantom.py"
    phantom_path = "./phantom.py"

    print(
        "This example uses the phantom definition from the package Tomograpy, "
        f"developed by Nicolas Barbey. The needed module will be downloaded from: {phantom_url}"
    )

    import urllib.request as urlreq

    urlreq.urlretrieve(phantom_url, phantom_path)

    with open(phantom_path, "r", encoding="utf-8") as f:
        file_content = f.read()
    with open(phantom_path, "w", encoding="utf-8") as f:
        f.write(file_content.replace("xrange", "range"))


def create_phantom_nuclei3d(
    FoV_size: Union[int, None] = 100, dtype: DTypeLike = np.float32
) -> tuple[NDArrayFloat, NDArrayFloat, NDArrayFloat]:
    """Create a 3D phantom of cell nuclei.

    Parameters
    ----------
    FoV_size : int | None
        Size of the field-of-view in pixels, per edge, by default None
    dtype : DTypeLike, optional
        The dtype of the produced data, by default np.float32

    Returns
    -------
    tuple[NDArrayFloat, NDArrayFloat, NDArrayFloat]
        The nuclei 3D phantom, the background, and the pixel sizes along each axis
    """
    phantom = np.array(skd.cells3d(), dtype=dtype).swapaxes(0, 1)
    nuclei: NDArrayFloat = phantom[1]

    if FoV_size is None:
        FoV_size = nuclei.shape[-1]
    else:
        nuclei = skt.rescale(nuclei, scale=FoV_size / nuclei.shape[-1], mode="reflect", channel_axis=0)
    print(f"{FoV_size = }")
    scale_factor = FoV_size / nuclei.shape[-1]
    pixel_size_um = np.array([0.29, 0.26, 0.26], dtype=dtype) / scale_factor

    circ_mask = processing.circular_mask([FoV_size, FoV_size], dtype=dtype)

    altitude = skf.sobel(nuclei)

    seeds_precip = nuclei > ((nuclei.max() - nuclei.min()) * 0.8 + nuclei.min())
    seeds_nuclei = nuclei > (nuclei.mean() + nuclei.std())

    seeds = np.zeros_like(nuclei, dtype=int)
    seeds[seeds_nuclei] = 2
    seeds[nuclei == nuclei.min()] = 1
    seg_nuclei = sks.watershed(altitude, seeds)
    seeds = np.zeros_like(nuclei, dtype=int)
    seeds[np.logical_and(seeds_nuclei, np.logical_not(circ_mask))] = 2
    seeds[nuclei == nuclei.min()] = 1
    seg_nuclei_out = sks.watershed(altitude, seeds)
    seg_nuclei = np.logical_and(seg_nuclei == 2, np.logical_not(seg_nuclei_out == 2))
    seg_nuclei = skm.remove_small_objects(seg_nuclei)
    seg_nuclei = skm.binary_dilation(seg_nuclei, footprint=skm.ball(1.5))

    seeds = np.zeros_like(nuclei, dtype=int)
    seeds[seeds_precip] = 2
    seeds[nuclei == nuclei.min()] = 1
    seg_precip = sks.watershed(skf.sobel(seg_nuclei), seeds)
    seg_precip = skm.binary_dilation(seg_precip == 2)

    ph_seg = seg_nuclei.copy() * 2
    ph_seg[seg_precip] = 3
    ph_seg[skm.dilation(seg_nuclei_out == 2, footprint=skm.ball(2.25))] = 1

    background = nuclei * (ph_seg == 0)
    for _ in range(15):
        if np.all(background != 0):
            break
        dil_back = skm.dilation(nuclei * (ph_seg == 0), footprint=skm.ball(1))
        background[background == 0] = dil_back[background == 0]
    background = skf.gaussian(background, sigma=9)

    foreground = nuclei * (ph_seg == 2) + (ph_seg != 2) * background
    foreground = skf.gaussian(foreground, sigma=0.75)

    return foreground, background, pixel_size_um


def phantom_assign_concentration(
    ph_or: NDArrayFloat, element: str = "Ca", em_line: str = "KA", in_energy_keV: float = 20.0, voxel_size_um: float = 0.5
) -> tuple[NDArrayFloat, NDArrayFloat, NDArrayFloat]:
    """Build an XRF phantom.

    The created phantom has been used in:
    - N. Viganò and V. A. Solé, “Physically corrected forward operators for
    induced emission tomography: a simulation study,” Meas. Sci. Technol., no.
    Advanced X-Ray Tomography, pp. 1–26, Nov. 2017.

    Parameters
    ----------
    ph_or : NDArrayFloat
        The phases phantom map.
    element : str, optional
        Element symbol. The default is "Ca".
    em_line : str, optional
        Emission line. The default is "KA" (corresponding to K-alpha).
    in_energy_keV : float, optional
        Incoming beam energy in keV. The default is 20.0.

    Returns
    -------
    vol_fluo_yield : NDArrayFloat
        Voxel-wise fluorescence yields.
    vol_att_in : NDArrayFloat
        Voxel-wise attenuation at the incoming beam energy.
    vol_att_out : NDArrayFloat
        Voxel-wise attenuation at the emitted energy.
    """
    ph_air = ph_or < 0.1
    ph_FeO = 0.5 < ph_or
    ph_CaO = np.logical_and(0.25 < ph_or, ph_or < 0.5)
    ph_CaC = np.logical_and(0.1 < ph_or, ph_or < 0.25)

    conv_mm_to_cm = 1e-1
    conv_um_to_mm = 1e-3
    voxel_size_cm = voxel_size_um * conv_um_to_mm * conv_mm_to_cm  # cm to micron
    print(f"Sample size: [{ph_or.shape[0] * voxel_size_um} {ph_or.shape[1] * voxel_size_um}] um")

    phase_fractions = (ph_air, ph_FeO, ph_CaO, ph_CaC)
    phase_compound_names = ("Air, Dry (near sea level)", "Ferric Oxide", "Calcium Oxide", "Calcium Carbonate")

    phantom = physics.VolumeMaterial(phase_fractions, phase_compound_names, voxel_size_cm)

    out_energy_keV, vol_fluo_yield = phantom.get_fluo_production(element, in_energy_keV, em_line)

    vol_lin_att_in = phantom.get_attenuation(in_energy_keV)
    vol_lin_att_out = phantom.get_attenuation(out_energy_keV)

    return (vol_fluo_yield, vol_lin_att_in, vol_lin_att_out)


def phantom_assign_concentration_multi(
    ph_or: NDArrayFloat,
    elements: Sequence[str] = ("Ca", "Fe"),
    em_lines: Union[str, Sequence[str]] = "KA",
    in_energy_keV: float = 20.0,
    detectors_pos_rad: Optional[float] = None,
) -> tuple[list[NDArrayFloat], NDArrayFloat, list[NDArrayFloat]]:
    """Build an XRF phantom.

    The created phantom has been used in:
    - N. Viganò and V. A. Solé, “Physically corrected forward operators for
    induced emission tomography: a simulation study,” Meas. Sci. Technol., no.
    Advanced X-Ray Tomography, pp. 1–26, Nov. 2017.

    Parameters
    ----------
    ph_or : NDArrayFloat
        The phases phantom map.
    elements : Sequence[str], optional
        Element symbols. The default is ["Ca", "Fe"].
    em_lines : str | Sequence[str], optional
        Emission lines. The default is "KA" (corresponding to K-alpha).
    in_energy_keV : float, optional
        Incoming beam energy in keV. The default is 20.0.
    detectors_pos_rad : float | tuple | list | NDArrayFloat, optional
        Detector(s) positions in radians, with respect to incoming beam.
        If None, Compton is not produced. The default is None.

    Returns
    -------
    vol_yield : List[NDArrayFloat]
        Voxel-wise fluorescence and Compton yields.
    vol_att_in : NDArrayFloat
        Voxel-wise attenuation at the incoming beam energy.
    vol_att_out : List[NDArrayFloat]
        Voxel-wise attenuation at the emitted energy.
    """
    ph_air = ph_or < 0.1
    ph_FeO = 0.5 < ph_or
    ph_CaO = np.logical_and(0.25 < ph_or, ph_or < 0.5)
    ph_CaC = np.logical_and(0.1 < ph_or, ph_or < 0.25)

    conv_mm_to_cm = 1e-1
    conv_um_to_mm = 1e-3
    voxel_size_um = 0.5
    voxel_size_cm = voxel_size_um * conv_um_to_mm * conv_mm_to_cm  # cm to micron
    print(f"Sample size: [{ph_or.shape[0] * voxel_size_um} {ph_or.shape[1] * voxel_size_um}] um")

    phase_fractions = (ph_air, ph_FeO, ph_CaO, ph_CaC)
    phase_compound_names = ("Air, Dry (near sea level)", "Ferric Oxide", "Calcium Oxide", "Calcium Carbonate")

    phantom = physics.VolumeMaterial(phase_fractions, phase_compound_names, voxel_size_cm)

    vol_lin_att_in = phantom.get_attenuation(in_energy_keV)

    num_vols_out = len(elements) + (detectors_pos_rad is not None)
    vol_yield = [np.array([])] * num_vols_out
    vol_lin_att_out = [np.array([])] * num_vols_out

    for ii, el in enumerate(elements):
        if isinstance(em_lines, str):
            line = em_lines
        else:
            line = em_lines[ii]
        out_energy_keV, vol_yield[ii] = phantom.get_fluo_production(el, in_energy_keV, line)
        vol_lin_att_out[ii] = phantom.get_attenuation(out_energy_keV)

    if detectors_pos_rad:
        out_energy_keV, vol_yield[-1] = phantom.get_compton_scattering(in_energy_keV, angle_rad=detectors_pos_rad)
        vol_lin_att_out[-1] = phantom.get_attenuation(out_energy_keV)

    return (vol_yield, vol_lin_att_in, vol_lin_att_out)


def add_noise(
    img_clean: NDArray,
    num_photons: Union[int, float],
    add_poisson: bool = False,
    readout_noise_std: Optional[float] = None,
    background_avg: Optional[float] = None,
    background_std: Optional[float] = None,
    detection_efficiency: float = 1.0,
    dtype: DTypeLike = np.float32,
) -> tuple[NDArray, NDArray, float]:
    """Add noise to an image (sinogram).

    Parameters
    ----------
    img_clean : NDArray
        The clean input image.
    num_photons : Union[int, float]
        Number of photons corresponding to the value 1.0 in the image.
    add_poisson : bool, optional
        Whether to add Poisson noise, by default False.
    readout_noise_std : Optional[float], optional
        Standard deviation of the readout noise, by default None.
    background_avg : Optional[float], optional
        Average value of the background, by default None.
    background_std : Optional[float], optional
        Standard deviation of the background, by default None.
    detection_efficiency : float, optional
        Efficiency of the detection (e.g. detector solid angle, inclination, etc), by default 1.0.
    dtype : DTypeLike, optional
        Data type of the volumes, by default np.float32.

    Returns
    -------
    Tuple[NDArray, NDArray, float]
        The noised and clean images (scaled by the photons and efficiency), and the background.
    """
    img_clean = num_photons * detection_efficiency * img_clean.copy().astype(dtype)

    img_noise = img_clean.copy()

    if background_avg is not None:
        if background_std is None:
            background_std = background_avg * 5e-2
        img_noise += (
            num_photons * detection_efficiency * np.abs(np.random.normal(background_avg, background_std, img_clean.shape))
        )

    background = float(np.mean((img_noise - img_clean).flatten()))

    if add_poisson:
        img_noise = np.random.poisson(img_noise).astype(dtype)
    if readout_noise_std is not None:
        img_noise += np.random.normal(0, readout_noise_std, img_clean.shape)

    return img_noise, img_clean, background


def create_sino(
    ph: NDArrayFloat,
    num_angles: int,
    start_angle_deg: float = 0.0,
    end_angle_deg: float = 180.0,
    dwell_time_s: float = 1.0,
    photon_flux: float = 1e9,
    detectors_pos_rad: Union[float, Sequence[float], NDArrayFloat] = (np.pi / 2),
    vol_att_in: Optional[NDArrayFloat] = None,
    vol_att_out: Optional[NDArrayFloat] = None,
    psf: Optional[NDArrayFloat] = None,
    background_avg: Optional[float] = None,
    background_std: Optional[float] = None,
    add_poisson: bool = False,
    readout_noise_std: Optional[float] = None,
    dtype: DTypeLike = np.float32,
) -> tuple[NDArrayFloat, NDArrayFloat, NDArrayFloat, float]:
    """Compute the sinogram from a given phantom.

    Parameters
    ----------
    ph : NDArrayFloat
        The phantom volume, with the expected average photon production per voxel per impinging photon.
    num_angles : int
        The number of angles.
    start_angle_deg : float, optional
        Initial scan angle in degrees. The default is 0.
    end_angle_deg : float, optional
        Final scan angle in degrees. The default is 180.
    dwell_time_s : float, optional
        The acquisition time per sinogram point. The default is 1.
    photon_flux : float, optional
        The impinging photon flux per unit time (second). The default is 1e9.
    detectors_pos_rad : float | Sequence[float] | NDArrayFloat, optional
        Detector(s) positions in radians, with respect to incoming beam. The default is (np.pi / 2).
    vol_att_in : NDArrayFloat, optional
        Attenuation volume for the incoming beam. The default is None.
    vol_att_out : NDArrayFloat, optional
        Attenuation volume for the outgoing beam. The default is None.
    psf : NDArrayFloat, optional
        Point spread function or probing beam profile. The default is None.
    background_avg : float, optional
        Background average value. The default is None.
    background_std : float, optional
        Background standard deviation. The default is None.
    add_poisson : bool, optional
        Switch to turn on Poisson noise. The default is False.
    readout_noise_std : float, optional
        Read-out noise standard deviation. The default is None.
    dtype : numpy.dtype, optional
        Output datatype. The default is np.float32.

    Returns
    -------
    Tuple[NDArrayFloat, NDArrayFloat, NDArrayFloat, float]
        The sinogram (detector readings), the angular positions, and the expected average phton production per voxel.
    """
    print(f"Creating Sino with {num_angles} angles")
    angles_deg = np.linspace(start_angle_deg, end_angle_deg, num_angles, endpoint=False)
    angles_rad = np.deg2rad(angles_deg)
    print(angles_deg)

    num_photons = photon_flux * dwell_time_s
    detector_solidangle_sr = (2.4 / 2 / 16 / 2) ** 2

    if vol_att_in is None and vol_att_out is None:
        with projectors.ProjectorUncorrected(ph.shape, angles_rad) as p:
            sino = p.fp(ph)
    else:
        with projectors.ProjectorAttenuationXRF(
            ph.shape, angles_rad, att_in=vol_att_in, att_out=vol_att_out, angles_detectors_rad=detectors_pos_rad, psf=psf
        ) as p:
            sino = p.fp(ph)

    # Adding noise
    sino_noise, sino, background = add_noise(
        sino,
        num_photons=num_photons,
        add_poisson=add_poisson,
        readout_noise_std=readout_noise_std,
        background_avg=background_avg,
        background_std=background_std,
        detection_efficiency=detector_solidangle_sr,
        dtype=dtype,
    )

    return (sino_noise, angles_rad, ph * num_photons * detector_solidangle_sr, background)


def create_sino_transmission(
    ph: NDArrayFloat,
    num_angles: int,
    start_angle_deg: float = 0,
    end_angle_deg: float = 180,
    dwell_time_s: float = 1,
    photon_flux: float = 1e9,
    psf: Optional[NDArrayFloat] = None,
    add_poisson: bool = False,
    readout_noise_std: Optional[float] = None,
    dtype: DTypeLike = np.float32,
) -> tuple[NDArrayFloat, NDArrayFloat, NDArrayFloat, NDArrayFloat]:
    """Compute the sinogram from a given phantom.

    Parameters
    ----------
    ph : NDArrayFloat
        The phantom volume, with the expected average photon production per voxel per impinging photon.
    num_angles : int
        The number of angles.
    start_angle_deg : float, optional
        Initial scan angle in degrees. The default is 0.
    end_angle_deg : float, optional
        Final scan angle in degrees. The default is 180.
    dwell_time_s : float, optional
        The acquisition time per sinogram point. The default is 1.
    photon_flux : float, optional
        The impinging photon flux per unit time (second). The default is 1e9.
    psf : NDArrayFloat, optional
        Point spread function or probing beam profile. The default is None.
    add_poisson : bool, optional
        Switch to turn on Poisson noise. The default is False.
    readout_noise_std : float, optional
        Read-out noise standard deviation. The default is None.
    dtype : numpy.dtype, optional
        Output datatype. The default is np.float32.

    Returns
    -------
    Tuple[NDArrayFloat, NDArrayFloat, NDArrayFloat, NDArrayFloat, float]
        The sinogram (detector readings), the flat-field, and the angular positions.
    """
    print(f"Creating attenuation Sino with {num_angles} angles")
    angles_deg = np.linspace(start_angle_deg, end_angle_deg, num_angles, endpoint=False)
    angles_rad = np.deg2rad(angles_deg)
    print(angles_deg)

    num_photons = photon_flux * dwell_time_s

    with projectors.ProjectorUncorrected(ph.shape, angles_rad, psf=psf) as p:
        sino = np.exp(-p.fp(ph))

    # Adding noise
    sino_noise, _, _ = add_noise(
        sino,
        num_photons=num_photons,
        add_poisson=add_poisson,
        readout_noise_std=readout_noise_std,
        dtype=dtype,
    )
    flat_noise, _, _ = add_noise(
        np.ones_like(sino),
        num_photons=num_photons,
        add_poisson=add_poisson,
        readout_noise_std=readout_noise_std,
        dtype=dtype,
    )

    return (sino_noise, flat_noise, angles_rad, ph)


def compute_error_power(expected_vol: NDArrayFloat, computed_vol: NDArrayFloat) -> tuple[float, float]:
    """Compute the expected volume signal power, and computed volume error power.

    Parameters
    ----------
    expected_vol : NDArrayFloat
        The expected volume.
    computed_vol : NDArrayFloat
        The computed volume.

    Returns
    -------
    Tuple[float, float]
        The expected volume signal power, and the computed volume.
    """
    vol_power = np.sqrt(np.sum((expected_vol) ** 2) / expected_vol.size)
    error_power = np.sqrt(np.sum(np.abs(expected_vol - computed_vol) ** 2) / expected_vol.size)
    return vol_power, error_power



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Data fidelity classes.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import numpy as np

from typing import Sequence, Union, Any
from numpy.typing import NDArray

from abc import ABC, abstractmethod

from . import operators

from copy import deepcopy


eps = np.finfo(np.float32).eps


NDArrayFloat = NDArray[np.floating]


def _soft_threshold(values: NDArrayFloat, threshold: Union[float, NDArrayFloat]) -> None:
    abs_values = np.abs(values)
    valid_values = abs_values > 0
    if isinstance(threshold, (float, int)) or threshold.size == 1:
        local_threshold = threshold
    else:
        local_threshold = threshold[valid_values]
    values[valid_values] *= np.fmax((abs_values[valid_values] - local_threshold) / abs_values[valid_values], 0)


class DataFidelityBase(ABC):
    """Define the DataFidelity classes interface."""

    data: Union[NDArrayFloat, None]
    sigma: Union[float, NDArrayFloat]
    background: Union[NDArrayFloat, None]

    sigma_data: Union[NDArrayFloat, None]

    __data_fidelity_name__ = ""

    def __init__(self, background: Union[float, NDArrayFloat, None] = None) -> None:
        """
        Initialize the base data-fidelity class.

        Parameters
        ----------
        background : float | NDArrayFloat | None, optional
            The data background. The default is None.
        """
        self.background = np.array(background) if background is not None else None
        self.data = None
        self.sigma = 1.0
        self.sigma_data = None

    def _slice_attr(self, attr: str, ind: Any) -> None:
        attr_val = self.__getattribute__(attr)
        if attr_val is not None and isinstance(attr_val, np.ndarray) and attr_val.size > 1:
            self.__setattr__(attr, attr_val[ind])

    def __getitem__(self, ind: Any) -> "DataFidelityBase":
        """
        Slice the norm and all its attributes.

        Parameters
        ----------
        ind : Any
            Slicing indices.

        Returns
        -------
        DataFidelityBase
            The sliced norm.
        """
        new_self = deepcopy(self)
        for attr in self.__dict__.keys():
            new_self._slice_attr(attr, ind)
        return new_self

    def info(self) -> str:
        """
        Return the data-fidelity info.

        Returns
        -------
        str
            Data fidelity info string.
        """
        if self.background is not None:
            if np.array(self.background).size > 1:
                bckgrnd_str = "(B:<array>)"
            else:
                bckgrnd_str = "(B:%g)" % self.background
        else:
            bckgrnd_str = ""
        return self.__data_fidelity_name__ + bckgrnd_str

    def upper(self) -> str:
        """
        Return the upper case name of the data-fidelity.

        Returns
        -------
        str
            Upper case string name of the data-fidelity.
        """
        return self.info().upper()

    def lower(self) -> str:
        """
        Return the lower case name of the data-fidelity.

        Returns
        -------
        str
            Lower case string name of the data-fidelity.
        """
        return self.info().lower()

    def assign_data(self, data: Union[float, NDArrayFloat, None] = None, sigma: Union[float, NDArrayFloat] = 1.0) -> None:
        """Initialize the data bias, and sigma of the data term.

        Parameters
        ----------
        data : Union[float, NDArrayFloat, None], optional
            The data bias, by default None
        sigma : Union[float, NDArrayFloat], optional
            The sigma, by default 1.0
        """
        self.data = np.array(data) if data is not None else None
        self.sigma = sigma
        self.sigma_data = self._compute_sigma_data()

    def compute_residual(self, proj_primal: NDArrayFloat, mask: Union[NDArrayFloat, None] = None) -> NDArrayFloat:
        """Compute the residual in the dual domain.

        Parameters
        ----------
        proj_primal : NDArrayFloat
            Projection of the primal solution
        mask : Union[NDArrayFloat, None], optional
            Mask of the dual domain, by default None

        Returns
        -------
        NDArrayFloat
            The residual
        """
        if self.background is not None:
            proj_primal = proj_primal + self.background

        if self.data is not None:
            residual = self.data - proj_primal
        else:
            residual = proj_primal.copy()

        if mask is not None:
            residual *= mask
        return residual

    @abstractmethod
    def compute_residual_norm(self, dual: NDArrayFloat) -> float:
        """Compute the norm of the residual.

        Parameters
        ----------
        dual : NDArrayFloat
            The residual in the dual domain.

        Returns
        -------
        float
            The residual norm.
        """

    def _compute_sigma_data(self):
        if self.data is None:
            return None
        else:
            return self.sigma * self.data

    def compute_data_dual_dot(self, dual: NDArrayFloat, mask: Union[NDArrayFloat, None] = None) -> float:
        """Compute the dot product of the data bias and the dual solution.

        Parameters
        ----------
        dual : NDArrayFloat
            The dual solution.
        mask : Union[NDArrayFloat, None], optional
            Mask of the dual domain, by default None

        Returns
        -------
        float
            The dot product between the data bias and the dual solution
        """
        if self.data is not None:
            if mask is not None:
                dual = dual * mask

            return np.dot(dual.flatten(), self.data.flatten())
        else:
            return 0.0

    def initialize_dual(self) -> NDArrayFloat:
        """Initialize the dual domain solution.

        Returns
        -------
        NDArrayFloat
            A zero array with the dimensions of the dual domain.
        """
        return np.zeros_like(self.data)

    def update_dual(self, dual: NDArrayFloat, proj_primal: NDArrayFloat) -> None:
        """Update the dual solution.

        Parameters
        ----------
        dual : NDArrayFloat
            The current dual solution
        proj_primal : NDArrayFloat
            The projected primal solution
        """
        if self.background is None:
            dual += proj_primal * self.sigma
        else:
            dual += (proj_primal + self.background) * self.sigma

    @abstractmethod
    def apply_proximal(self, dual: NDArrayFloat) -> None:
        """Apply the proximal in the dual domain.

        Parameters
        ----------
        dual : NDArrayFloat
            The dual solution
        """

    @abstractmethod
    def compute_primal_dual_gap(
        self, proj_primal: NDArrayFloat, dual: NDArrayFloat, mask: Union[NDArrayFloat, None] = None
    ) -> float:
        """Compute the primal-dual gap of the current solution.

        Parameters
        ----------
        proj_primal : NDArrayFloat
            The projected primal solution (in the dual domain)
        dual : NDArrayFloat
            The dual solution
        mask : Union[NDArrayFloat, None], optional
            Mask in the dual domain, by default None

        Returns
        -------
        float
            The primal-dual gap
        """


class DataFidelity_l2(DataFidelityBase):
    """l2-norm data-fidelity class."""

    __data_fidelity_name__ = "l2"

    def __init__(self, background: Union[float, NDArrayFloat, None] = None) -> None:
        super().__init__(background=background)

    def assign_data(self, data: Union[float, NDArrayFloat, None] = None, sigma: Union[float, NDArrayFloat] = 1.0) -> None:
        super().assign_data(data=data, sigma=sigma)
        self.sigma1 = 1 / (1 + sigma)

    def compute_residual_norm(self, dual: NDArrayFloat) -> float:
        return float(np.linalg.norm(dual.flatten(), ord=2) ** 2)

    def apply_proximal(self, dual: NDArrayFloat) -> None:
        if self.data is not None and self.sigma_data is not None:
            dual -= self.sigma_data
        dual *= self.sigma1

    def compute_primal_dual_gap(
        self, proj_primal: NDArrayFloat, dual: NDArrayFloat, mask: Union[NDArrayFloat, None] = None
    ) -> float:
        return float(
            np.linalg.norm(self.compute_residual(proj_primal, mask), ord=2) + np.linalg.norm(dual, ord=2)
        ) / 2 + self.compute_data_dual_dot(dual)


class DataFidelity_wl2(DataFidelity_l2):
    """Weighted l2-norm data-fidelity class."""

    __data_fidelity_name__ = "wl2"

    def __init__(self, weights: Union[float, NDArrayFloat], background: Union[float, NDArrayFloat, None] = None) -> None:
        super().__init__(background=background)
        self.weights = np.array(weights)

    def assign_data(self, data: Union[float, NDArrayFloat, None], sigma: Union[float, NDArrayFloat] = 1.0):
        super().assign_data(data=data, sigma=sigma)
        if isinstance(self.sigma, np.ndarray):
            dtype = self.sigma.dtype
        else:
            dtype = type(self.sigma)
        invalid_weights = (self.weights == 0).astype(dtype)
        self.sigma1 = 1 / (1 + sigma / (self.weights + invalid_weights)) * (1 - invalid_weights)

    def compute_residual(self, proj_primal, mask: Union[float, NDArrayFloat, None] = None):
        if self.background is not None:
            proj_primal = proj_primal + self.background
        if self.data is not None:
            residual = (self.data - proj_primal) * self.weights
        else:
            residual = proj_primal * self.weights
        if mask is not None:
            residual *= mask
        return residual

    def compute_residual_norm(self, dual: Union[float, NDArrayFloat]) -> float:
        valid_weights = self.weights != 0
        if isinstance(dual, np.ndarray):
            dual = dual[valid_weights]
        weights = self.weights[valid_weights]
        return float(np.linalg.norm((dual / np.sqrt(weights)).flatten(), ord=2) ** 2)


class DataFidelity_l2b(DataFidelity_l2):
    """l2-norm ball data-fidelity class."""

    __data_fidelity_name__ = "l2b"

    def __init__(self, local_error: Union[float, NDArrayFloat], background: Union[float, NDArrayFloat, None] = None):
        super().__init__(background=background)
        self.local_error = local_error

    def assign_data(self, data: Union[float, NDArrayFloat, None], sigma: Union[float, NDArrayFloat] = 1.0):
        self.sigma_error = sigma * self.local_error
        self.sigma_sqrt_error = sigma * np.sqrt(self.local_error)
        super().assign_data(data=data, sigma=sigma)
        self.sigma1 = 1 / (1 + self.sigma_error)

    def compute_residual(self, proj_primal: NDArrayFloat, mask: Union[NDArrayFloat, None] = None) -> NDArrayFloat:
        residual = super().compute_residual(proj_primal, mask)
        _soft_threshold(residual, self.sigma_sqrt_error)
        return residual

    def apply_proximal(self, dual: NDArrayFloat) -> None:
        if self.data is not None and self.sigma_data is not None:
            dual -= self.sigma_data
        _soft_threshold(dual, self.sigma_sqrt_error)
        dual *= self.sigma1

    def compute_primal_dual_gap(
        self, proj_primal: NDArrayFloat, dual: NDArrayFloat, mask: Union[NDArrayFloat, None] = None
    ) -> float:
        return float(
            np.linalg.norm(self.compute_residual(proj_primal, mask), ord=2)
            + np.linalg.norm(np.sqrt(self.local_error) * dual, ord=2)
        ) / 2 + self.compute_data_dual_dot(dual)


class DataFidelity_Huber(DataFidelityBase):
    """Huber-norm data-fidelity class. Given a parameter a: l2-norm for x < a, and l1-norm for x > a."""

    __data_fidelity_name__ = "Hub"

    def __init__(self, local_error, background=None, l2_axis=None):
        super().__init__(background=background)
        self.local_error = local_error
        self.l2_axis = l2_axis

    def assign_data(self, data, sigma=1.0):
        self.one_sigma_error = 1 / (1 + sigma * self.local_error)
        super().assign_data(data=data, sigma=sigma)

    def compute_residual_norm(self, dual):
        l2_points = dual <= self.local_error
        l1_points = 1 - l2_points
        return np.linalg.norm(dual[l2_points].flatten(), ord=2) ** 2 + np.linalg.norm(dual[l1_points].flatten(), ord=1)

    def apply_proximal(self, dual):
        if self.data is not None and self.sigma_data is not None:
            dual -= self.sigma_data

        dual *= self.one_sigma_error

        if self.l2_axis is None:
            dual /= np.fmax(1, np.abs(dual))
        else:
            dual_dir_norm_l2 = np.linalg.norm(dual, ord=2, axis=self.l2_axis, keepdims=True)
            dual /= np.fmax(1, dual_dir_norm_l2)

    def compute_primal_dual_gap(self, proj_primal, dual, mask=None):
        if self.background is not None:
            proj_primal = proj_primal + self.background
        return (
            np.linalg.norm(self.compute_residual(proj_primal, mask), ord=2)
            + self.compute_data_dual_dot(dual)
            + self.local_error * np.linalg.norm(dual, ord=2)
        )


class DataFidelity_l1(DataFidelityBase):
    """l1-norm data-fidelity class."""

    __data_fidelity_name__ = "l1"

    def __init__(self, background=None):
        super().__init__(background=background)

    def _get_inner_norm(self, dual):
        return np.abs(dual)

    def _apply_threshold(self, dual):
        pass

    def apply_proximal(self, dual, weight=1.0):
        if self.data is not None:
            dual -= self.sigma_data
        self._apply_threshold(dual)
        dual_inner_norm = self._get_inner_norm(dual)
        dual /= np.fmax(dual_inner_norm, weight)
        dual *= weight

    def compute_residual_norm(self, dual):
        dual = dual.copy()
        self._apply_threshold(dual)
        dual_inner_norm = self._get_inner_norm(dual)
        return np.linalg.norm(dual_inner_norm, ord=1)

    def compute_primal_dual_gap(self, proj_primal, dual, mask=None):
        if self.background is not None:
            proj_primal = proj_primal + self.background

        residual = self.compute_residual(proj_primal, mask)
        self._apply_threshold(residual)
        residual_inner_norm = self._get_inner_norm(residual)
        return np.linalg.norm(residual_inner_norm, ord=1) + self.compute_data_dual_dot(dual)


class DataFidelity_l12(DataFidelity_l1):
    """l12-norm data-fidelity class."""

    __data_fidelity_name__ = "l12"

    def __init__(self, background=None, l2_axis=0):
        super().__init__(background=background)
        self.l2_axis = l2_axis

    def _get_inner_norm(self, dual):
        return np.linalg.norm(dual, ord=2, axis=self.l2_axis, keepdims=True)


class DataFidelity_l1b(DataFidelity_l1):
    """l1-norm ball data-fidelity class."""

    __data_fidelity_name__ = "l1b"

    def __init__(self, local_error, background=None):
        super().__init__(background=background)
        self.local_error = local_error

    def assign_data(self, data, sigma=1.0):
        self.sigma_error = sigma * self.local_error
        super().assign_data(data=data, sigma=sigma)

    def _apply_threshold(self, dual):
        _soft_threshold(dual, self.local_error)


class DataFidelity_KL(DataFidelityBase):
    """KullbackLeibler data-fidelity class."""

    __data_fidelity_name__ = "KL"

    def _compute_sigma_data(self):
        if self.data is None:
            return None
        else:
            return 4 * self.sigma * np.fmax(self.data, 0.0)

    def apply_proximal(self, dual):
        if self.sigma_data is not None:
            dual[:] = (1 + dual[:] - np.sqrt((dual[:] - 1) ** 2 + self.sigma_data[:])) / 2
        else:
            dual[:] = (1 + dual[:] - np.sqrt((dual[:] - 1) ** 2)) / 2

    def compute_residual(self, proj_primal, mask=None):
        if self.background is not None:
            proj_primal = proj_primal + self.background
        # we take the Moreau envelope here, and apply the proximal to it
        residual = np.fmax(proj_primal, eps) * self.sigma

        self.apply_proximal(residual)

        if mask is not None:
            residual *= mask
        return -residual

    def compute_residual_norm(self, dual):
        return np.linalg.norm(dual.flatten(), ord=1)

    def compute_primal_dual_gap(self, proj_primal, dual, mask=None):
        if self.background is not None:
            proj_primal = proj_primal + self.background

        if self.data is not None:
            data_nn = np.fmax(self.data, eps)
            proj_primal_nn = np.fmax(proj_primal, eps)
            residual = proj_primal_nn - data_nn * (1 - np.log(data_nn) + np.log(proj_primal_nn))
        else:
            residual = np.copy(proj_primal)
        if mask is not None:
            residual *= mask

        return np.linalg.norm(residual, ord=1)


class DataFidelity_ln(DataFidelityBase):
    """nuclear-norm data-fidelity class."""

    __data_fidelity_name__ = "ln"

    def __init__(self, background=None, ln_axes: Sequence[int] = (1, -1), spectral_norm: DataFidelityBase = DataFidelity_l1()):
        super().__init__(background=background)
        self.ln_axes = ln_axes
        self.spectral_norm = spectral_norm
        self.use_fallback = False

    def apply_proximal(self, dual):
        dual_tmp = dual.copy()

        if self.sigma_data is not None:
            # If we have a bias term, we interpret it as an addition to the rows in the SVD decomposition.
            # Performing this operation before the transpose is a waste of computation, but it simplifies the logic.
            dual_tmp = np.concatenate((dual_tmp, self.sigma_data), axis=self.ln_axes[0])

        if self.use_fallback:
            t_range = [*range(len(dual_tmp.shape))]
            t_range.append(t_range.pop(self.ln_axes[0]))
            t_range.append(t_range.pop(self.ln_axes[1]))
            dual_tmp = np.transpose(dual_tmp, t_range)

            U, s_p, Vt = np.linalg.svd(dual_tmp, full_matrices=False)

            self.spectral_norm.apply_proximal(s_p)

            dual_tmp = np.matmul(U, s_p[..., None] * Vt)
            dual_tmp = np.transpose(dual_tmp, np.argsort(t_range))
        else:
            op_svd = operators.TransformSVD(dual_tmp.shape, axes_rows=self.ln_axes[0], axes_cols=self.ln_axes[1])
            s_p = op_svd(dual_tmp)

            self.spectral_norm.apply_proximal(s_p)

            dual_tmp = op_svd.T(s_p)

        if self.data is not None:
            # We now strip the bias data, to make sure that we don't change dimensionality.
            # dual_tmp = dual_tmp[..., : dual_tmp.shape[-2] - 1 :, :]
            dual_tmp = np.take(dual_tmp, np.arange(dual_tmp.shape[self.ln_axes[0]] - 1), axis=self.ln_axes[0])

        dual[:] = dual_tmp[:]

    def compute_residual_norm(self, dual):
        op_svd = operators.TransformSVD(dual.shape, axes_rows=self.ln_axes[0], axes_cols=self.ln_axes[1])
        s_p = op_svd(dual)
        return np.linalg.norm(s_p, ord=1)

    def compute_primal_dual_gap(self, proj_primal, dual, mask=None):
        if self.background is not None:
            proj_primal = proj_primal + self.background

        residual = self.compute_residual(proj_primal, mask)

        return self.compute_residual_norm(residual) + self.compute_data_dual_dot(dual)



"""Physics module."""

__author__ = """Nicola VIGANÒ"""
__email__ = "N.R.Vigano@cwi.nl"


import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import NDArray

from . import attenuation  # noqa: F401, F402
from . import materials  # noqa: F401, F402
from . import phase  # noqa: F401, F402
from . import xraylib_helper  # noqa: F401, F402
from . import xrf  # noqa: F401, F402
from . import units

xraylib = xraylib_helper.xraylib
get_compound = xraylib_helper.get_compound
get_element_number = xraylib_helper.get_element_number

FluoLinesSiegbahn = xrf.LinesSiegbahn
VolumeMaterial = materials.VolumeMaterial

convert_energy_to_wlength = units.energy_to_wlength
convert_wlength_to_energy = units.wlength_to_energy


def pencil_beam_profile(
    voxel_size_um: float, beam_fwhm_um: float, profile_size: int = 1, beam_shape: str = "gaussian", verbose: bool = False
) -> NDArray:
    """
    Compute the pencil beam integration point spread function.

    Parameters
    ----------
    voxel_size_um : float
        The integration length.
    beam_fwhm_um : float
        The beam FWHM.
    profile_size : int, optional
        The number of pixels of the PSF. The default is 1.
    verbose : bool, optional
        Whether to print verbose information. The default is False.

    Returns
    -------
    NDArray
        The beam PSF.
    """
    y_points = profile_size * 2 + 1
    extent_um = y_points * voxel_size_um
    num_points = int(np.ceil(extent_um * 10) // 2) * 2 + 1
    half_voxel_size_um = voxel_size_um / 2

    eps = np.finfo(np.float32).eps

    xc = np.linspace(-extent_um / 2, extent_um / 2, num_points)

    # Box beam shape
    yb = np.abs(xc) < half_voxel_size_um
    yb[np.abs(np.abs(xc) - half_voxel_size_um) <= eps] = 0.5

    # Gaussian beam shape
    if beam_shape.lower() == "gaussian":
        yg = np.exp(-4 * np.log(2) * (xc**2) / (beam_fwhm_um**2))
    elif beam_shape.lower() == "lorentzian":
        beam_hwhm_um = beam_fwhm_um / 2
        yg = (beam_hwhm_um**2) / (xc**2 + beam_hwhm_um**2)
    elif beam_shape.lower() == "sech**2":
        # doi: 10.1364/ol.20.001160
        tau = beam_fwhm_um / (2 * np.arccosh(np.sqrt(2)))
        yg = 1 / np.cosh(xc / tau) ** 2
    else:
        raise ValueError(f"Unknown beam shape: {beam_shape.lower()}")

    yc = np.convolve(yb, yg, mode="same")
    yc = yc / np.max(yc)

    y = np.zeros((y_points,))
    for ii_p in range(y_points):
        # Finding the region that overlaps with the given adjacent voxel
        voxel_center_um = (ii_p - profile_size) * voxel_size_um
        yp = np.abs(xc - voxel_center_um) < half_voxel_size_um
        yp[(np.abs(xc - voxel_center_um) - half_voxel_size_um) < eps] = 0.5

        y[ii_p] = np.sum(yc * yp)

    # Renormalization
    y /= y.sum()

    if verbose:
        fig, ax = plt.subplots(1, 2, figsize=[10, 5])
        ax[0].plot(xc, yb, label="Integration length")  # type: ignore
        ax[0].plot(xc, yg, label=f"{beam_shape.capitalize()} beam shape")  # type: ignore
        ax[0].plot(xc, yc, label="Resulting beam shape")  # type: ignore
        ax[0].legend()  # type: ignore
        ax[0].grid()  # type: ignore

        pixels_pos = np.linspace(-(y_points - 1) / 2, (y_points - 1) / 2, y_points)
        ax[1].bar(pixels_pos, y, label="PSF values", color="C1", width=1, linewidth=1.5, edgecolor="k", alpha=0.75)  # type: ignore
        ax[1].plot(xc / extent_um * profile_size * 2, yc, label="Resulting beam shape")  # type: ignore
        ax[1].legend()  # type: ignore
        ax[1].grid()  # type: ignore
        fig.tight_layout()

    return y



import numpy as np
import pytest
from src.corrct.struct_illum import (
    reorder_masks,
    decompose_qr_masks,
    estimate_resolution,
    MaskCollection,
    MaskGeneratorPoint,
    MaskGeneratorBernoulli,
    MaskGeneratorHalfGaussian,
    MaskGeneratorMURA,
    ProjectorGhostImaging,
)


def test_reorder_masks():
    """
    Test the reorder_masks function.

    This function tests the reordering of masks and buckets based on a given shift.
    """
    masks = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
    buckets = np.array([1, 2])
    shift = 1
    reordered_masks, reordered_buckets = reorder_masks(masks, buckets, shift)
    assert np.array_equal(reordered_masks, np.array([[[5, 6], [7, 8]], [[1, 2], [3, 4]]]))
    assert np.array_equal(reordered_buckets, np.array([2, 1]))


def test_decompose_qr_masks():
    """
    Test the decompose_qr_masks function.

    This function tests the decomposition of masks into Q and R1t matrices.
    """
    masks = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
    Q, R1t = decompose_qr_masks(masks)
    assert Q.shape == masks.shape
    assert R1t.shape == (masks.shape[0], masks.shape[0])


def test_mask_collection():
    """
    Test the MaskCollection class.

    This function tests the creation and properties of the MaskCollection class.
    """
    masks_enc = np.random.rand(2, 2, 2)
    masks_dec = np.random.rand(2, 2, 2)
    mc = MaskCollection(masks_enc, masks_dec)
    assert mc.shape_fov == (2, 2)
    assert mc.shape_shifts == (2,)
    assert mc.num_buckets == 2
    assert mc.num_pixels == 4


def test_mask_generator_point():
    """
    Test the MaskGeneratorPoint class.

    This function tests the creation and properties of the MaskGeneratorPoint class.
    """
    mg = MaskGeneratorPoint(fov_size_mm=10, req_res_mm=1)
    assert mg.shape_fov[0] == 10
    assert mg.shape_mask[0] == 1
    assert mg.shape_shifts[0] == 10


def test_mask_generator_bernoulli():
    """
    Test the MaskGeneratorBernoulli class.

    This function tests the creation and properties of the MaskGeneratorBernoulli class.
    """
    mg = MaskGeneratorBernoulli(fov_size_mm=10, req_res_mm=1)
    assert mg.shape_fov[0] == 10
    assert mg.shape_mask[0] == 10
    assert mg.shape_shifts[0] == 12


def test_mask_generator_half_gaussian():
    """
    Test the MaskGeneratorHalfGaussian class.

    This function tests the creation and properties of the MaskGeneratorHalfGaussian class.
    """
    mg = MaskGeneratorHalfGaussian(fov_size_mm=10, req_res_mm=1)
    assert mg.shape_fov[0] == 10
    assert mg.shape_mask[0] == 10
    assert mg.shape_shifts[0] == 12


def test_mask_generator_mura():
    """
    Test the MaskGeneratorMURA class.

    This function tests the creation and properties of the MaskGeneratorMURA class.
    """
    mg = MaskGeneratorMURA(fov_size_mm=10, req_res_mm=1)
    assert mg.shape_fov[0] == 13
    assert mg.shape_mask[0] == 13
    assert mg.shape_shifts[0] == 13


def test_projector_ghost_imaging():
    """
    Test the ProjectorGhostImaging class.

    This function tests the creation and properties of the ProjectorGhostImaging class,
    as well as its forward projection and least-squares reconstruction methods.
    """
    masks = np.random.rand(2, 2, 2)
    projector = ProjectorGhostImaging(masks)
    assert projector.mc.shape_fov == (2, 2)
    assert projector.mc.shape_shifts == (2,)
    assert projector.mc.num_buckets == 2
    assert projector.mc.num_pixels == 4

    image = np.random.rand(2, 2)
    bucket_vals = projector.fp(image)
    assert bucket_vals.shape == (2,)

    reconstructed_image = projector.fbp(bucket_vals)
    assert reconstructed_image.shape == (2, 2)


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
"""
Volume centering classes.

@author: Nicola VIGANÒ, CEA-IRIG and ESRF - The European Synchrotron, Grenoble, France
"""

import numpy as np

from typing import Union
from numpy.typing import ArrayLike, NDArray

from . import fitting

from .. import models
from ..processing import post as post_proc


class RecenterVolume:
    """Volume re-centering class."""

    def __init__(
        self, proj_geom: models.ProjectionGeometry, angles_rad: Union[NDArray, ArrayLike], precision: int = 2
    ) -> None:
        """Volume recentering class, that shifts the detector position, in order to meet a certain volume position.

        Parameters
        ----------
        proj_geom : models.ProjectionGeometry
            The projection geometry.
        angles_rad : NDArray | ArrayLike
            The rotation angles to consider.
        precision : int, optional
            Number of decimals to use, by default 2.
        """
        self.prj_geom = proj_geom.rotate(angles_rad)
        self.precision = precision

    def _apply_displacement_vu(self, shifts_vu: NDArray, displacemenet_zyx: NDArray) -> NDArray:
        shifts_vu_corrs = self.prj_geom.project_displacement_to_detector(displacemenet_zyx)
        return np.around(shifts_vu + shifts_vu_corrs, decimals=self.precision)

    def to_com(self, shifts_vu: Union[ArrayLike, NDArray], volume: NDArray, com_ref_zyx: Union[ArrayLike, NDArray]) -> NDArray:
        """Recenter to a given center-of-mass (CoM).

        Parameters
        ----------
        shifts_vu : ArrayLike | NDArray
            The current VU shifts.
        volume : NDArray
            The volume to shift.
        com_ref_zyx : ArrayLike | NDArray
            The destination CoM.

        Returns
        -------
        NDArray
            The corrected VU shifts.
        """
        com_rec_zyx = post_proc.com(volume)
        displacemenet_zyx = np.array(com_ref_zyx) - com_rec_zyx
        return self._apply_displacement_vu(np.array(shifts_vu), displacemenet_zyx)

    def as_reference(self, shifts_vu: NDArray, volume: NDArray, reference: NDArray, method: str = "com") -> NDArray:
        """Recenter with respect to a given volume.

        Parameters
        ----------
        shifts_vu : NDArray
            Current VU shifts.
        volume : NDArray
            The volume to shift.
        reference : NDArray
            The reference volume.
        method : str, optional
            The method to use out of "com" | "xc" (cross-correlation), by default "com"

        Returns
        -------
        NDArray
            The corrected VU shifts.

        Raises
        ------
        ValueError
            In case of wrong method.
        """
        if method.lower() == "com":
            com_ref_zyx = post_proc.com(reference)
            return self.to_com(shifts_vu, volume, com_ref_zyx)
        elif method.lower() == "xc":
            displacemenet_zyx = fitting.fit_shifts_zyx_xc(reference, volume, decimals=self.precision)
            return self._apply_displacement_vu(shifts_vu, displacemenet_zyx)
        else:
            raise ValueError(f"Method (passed: {method}) should be one of: 'com' | 'xc'.")



# test_models.py

import numpy as np
import pytest
from src.corrct.models import (
    ProjectionGeometry,
    VolumeGeometry,
    get_prj_geom_parallel,
    get_prj_geom_cone,
    get_vol_geom_from_data,
    get_vol_geom_from_volume,
    combine_shifts_vu,
    get_rot_axis_dir,
)


def test_projection_geometry_initialization():
    """
    Test the initialization of the ProjectionGeometry class.
    """
    geom = ProjectionGeometry(
        geom_type="parallel3d",
        src_pos_xyz=np.array([0.0, -1.0, 0.0]),
        det_pos_xyz=np.zeros(3),
        det_u_xyz=np.array([1.0, 0.0, 0.0]),
        det_v_xyz=np.array([0.0, 0.0, 1.0]),
        rot_dir_xyz=np.array([0.0, 0.0, -1.0]),
    )
    assert geom.geom_type == "parallel3d"
    assert np.array_equal(geom.src_pos_xyz, np.array([[0.0, -1.0, 0.0]]))
    assert np.array_equal(geom.det_pos_xyz, np.array([[0.0, 0.0, 0.0]]))
    assert np.array_equal(geom.det_u_xyz, np.array([[1.0, 0.0, 0.0]]))
    assert np.array_equal(geom.det_v_xyz, np.array([[0.0, 0.0, 1.0]]))
    assert np.array_equal(geom.rot_dir_xyz, np.array([[0.0, 0.0, -1.0]]))


def test_volume_geometry_initialization():
    """
    Test the initialization of the VolumeGeometry class.
    """
    vol_geom = VolumeGeometry(_vol_shape_xyz=np.array([100, 100, 50]), vox_size=1.0)
    assert np.array_equal(vol_geom.shape_xyz, np.array([100, 100, 50]))
    assert vol_geom.vox_size == 1.0


def test_get_prj_geom_parallel():
    """
    Test the generation of a parallel beam geometry.
    """
    geom = get_prj_geom_parallel(geom_type="3d", rot_axis_shift_pix=None, rot_axis_dir="clockwise")
    assert geom.geom_type == "parallel3d"
    assert np.array_equal(geom.src_pos_xyz, np.array([[0.0, -1.0, 0.0]]))
    assert np.array_equal(geom.det_pos_xyz, np.array([[0.0, 0.0, 0.0]]))
    assert np.array_equal(geom.det_u_xyz, np.array([[1.0, 0.0, 0.0]]))
    assert np.array_equal(geom.det_v_xyz, np.array([[0.0, 0.0, 1.0]]))
    assert np.array_equal(geom.rot_dir_xyz, np.array([[0.0, 0.0, -1.0]]))


def test_get_prj_geom_cone():
    """
    Test the generation of a cone beam geometry.
    """
    geom = get_prj_geom_cone(src_to_sam_dist=100.0, rot_axis_shift_pix=None, rot_axis_dir="clockwise")
    assert geom.geom_type == "cone"
    assert np.array_equal(geom.src_pos_xyz, np.array([[0.0, -100.0, 0.0]]))
    assert np.array_equal(geom.det_pos_xyz, np.array([[0.0, 0.0, 0.0]]))
    assert np.array_equal(geom.det_u_xyz, np.array([[1.0, 0.0, 0.0]]))
    assert np.array_equal(geom.det_v_xyz, np.array([[0.0, 0.0, 1.0]]))
    assert np.array_equal(geom.rot_dir_xyz, np.array([[0.0, 0.0, -1.0]]))


def test_get_vol_geom_from_data():
    """
    Test the generation of a volume geometry from data.
    """
    data = np.random.rand(10, 20, 30)
    vol_geom = get_vol_geom_from_data(data, data_format="dvwu")
    assert np.array_equal(vol_geom.shape_xyz, np.array([30, 30, 10]))


def test_get_vol_geom_from_volume():
    """
    Test the generation of a volume geometry from a volume.
    """
    volume = np.random.rand(10, 20, 30)
    vol_geom = get_vol_geom_from_volume(volume)
    # XXX: There might be a mess up here between X and Y!!
    assert np.array_equal(vol_geom.shape_xyz, np.array([20, 30, 10]))


def test_combine_shifts_vu():
    """
    Test the combination of vertical and horizontal shifts.
    """
    shifts_v = np.array([1, 2, 3])
    shifts_u = np.array([4, 5, 6])
    combined_shifts = combine_shifts_vu(shifts_v, shifts_u)
    assert np.array_equal(combined_shifts, np.array([[1, 2, 3], [4, 5, 6]]))

    # Test with invalid shapes
    with pytest.raises(ValueError, match="Expected 1-dimensional array for vertical shifts"):
        combine_shifts_vu(np.array([[1, 2], [3, 4]]), shifts_u)

    with pytest.raises(ValueError, match="Expected 1-dimensional array for horizontal shifts"):
        combine_shifts_vu(shifts_v, np.array([[4, 5], [6, 7]]))

    with pytest.raises(ValueError, match="Number of vertical shifts"):
        combine_shifts_vu(np.array([1, 2]), shifts_u)


def test_get_rot_axis_dir():
    """
    Test the processing of the rotation axis direction.
    """
    rot_dir = get_rot_axis_dir("clockwise")
    assert np.array_equal(rot_dir, np.array([0.0, 0.0, -1.0]))

    rot_dir = get_rot_axis_dir("counter-clockwise")
    assert np.array_equal(rot_dir, np.array([0.0, 0.0, 1.0]))

    rot_dir = get_rot_axis_dir(np.array([1.0, 0.0, 0.0]))
    assert np.array_equal(rot_dir, np.array([1.0, 0.0, 0.0]))

    # Test with invalid direction
    with pytest.raises(ValueError, match="Rotation axis direction invalid_dir not allowed"):
        get_rot_axis_dir("invalid_dir")
